import sys
from typing import Optional, Tuple, Union

from pytorch_lightning import Trainer
from pytorch_lightning.core.memory import ModelSummary
from pytorch_lightning.loggers.base import LightningLoggerBase
from pytorch_lightning.utilities.apply_func import move_data_to_device
import torch
from tqdm.auto import tqdm

from collie.model.base.base_pipeline import BasePipeline
from collie.model.base.layers import MultiLRScheduler, MultiOptimizer


class CollieTrainer(Trainer):
    """
    Helper wrapper class around PyTorch Lightning's ``Trainer`` class.

    Specifically, this wrapper:

    * Checks if a model has a validation dataset passed in (under the ``val_loader`` attribute)
      and, if not, sets ``num_sanity_val_steps`` to 0 and ``check_val_every_n_epoch`` to
      ``sys.maxint``.

    * Checks if a GPU is available and, if ``gpus is None``, sets ``gpus = -1``.

    See ``pytorch_lightning.Trainer`` documentation for more details at:
    https://pytorch-lightning.readthedocs.io/en/latest/common/trainer.html#trainer-class-api

    Compared with ``CollieMinimalTrainer``, PyTorch Lightning's ``Trainer`` offers more flexibility
    and room for exploration, at the cost of a higher training time (which is especially true for
    larger models). We recommend starting all model exploration with this ``CollieTrainer``
    (callbacks, automatic Lightning optimizations, etc.), finding a set of hyperparameters that work
    for your training job, then using this in the simpler but faster ``CollieMinimalTrainer``.

    Parameters
    ----------
    model: collie.model.BasePipeline
        Initialized Collie model
    benchmark: bool
        If set to ``True``, enables ``cudnn.benchmark``
    deterministic: bool
        If set to ``True``, enables ``cudnn.deterministic``
    **kwargs: keyword arguments
        Additional keyword arguments to be sent to the ``Trainer`` class:
        https://pytorch-lightning.readthedocs.io/en/latest/common/trainer.html#trainer-class-api

    """
    def __init__(self,
                 model: torch.nn.Module,
                 benchmark: bool = True,
                 deterministic: bool = True,
                 **kwargs):
        if not hasattr(model, 'val_loader') or model.val_loader is None:
            print('Did not detect ``val_loader``. Setting ``num_sanity_val_steps`` to 0.')
            kwargs['num_sanity_val_steps'] = 0
            kwargs['check_val_every_n_epoch'] = sys.maxsize

        if kwargs.get('gpus') is None and torch.cuda.is_available():
            print('Detected GPU. Setting ``gpus`` to 1.')
            kwargs['gpus'] = 1

        kwargs['benchmark'] = benchmark
        kwargs['deterministic'] = deterministic

        super().__init__(**kwargs)


class CollieMinimalTrainer():
    """
    A more manual implementation of PyTorch Lightning's ``Trainer`` class, attempting to port over
    the most commonly used ``Trainer`` arguments into a training loop with more transparency and
    faster training times.

    Through extensive experimentation, we found that PyTorch Lightning's ``Trainer`` was training
    Collie models about 25% slower than the more manual, typical PyTorch training loop boilerplate.
    Thus, we created the ``CollieMinimalTrainer``, which shares a similar API to PyTorch Lightning's
    ``Trainer`` object (both in instantiation and in usage), with a standard PyTorch training loop
    in its place.

    While PyTorch Lightning's ``Trainer`` offers more flexibility and customization through the
    addition of the additional ``Trainer`` arguments and ``callbacks``, we designed this class as a
    way to train a model in production, where we might be more focused on faster training times and
    less on hyperparameter tuning and R&D, where one might instead opt to use PyTorch Lightning's
    ``Trainer`` class.

    Note that the arguments the ``CollieMinimalTrainer`` trainer accepts will be slightly different
    than the ones that the ``CollieTrainer`` accept, and defaults are also not guaranteed to be
    equal as the two libraries evolve. Notable changes are:

    * If ``gpus > 1``, only a single GPU will be used and any other GPUs will remain unused. Multi-
      GPU training is not supported in ``CollieMinimalTrainer`` at this time.

    * ``logger == True`` has no meaning in ``CollieMinimalTrainer`` - a default logger will NOT be
      created if set to ``True``.

    * There is no way to pass in ``callbacks`` at this time. Instead, we will implement the most
      used ones during training here, manually, in favor of greater speed over customization.
      To use early stopping, set the ``early_stopping_patience`` to an integer other than ``None``.

    .. code-block:: python

        from collie.model import CollieMinimalTrainer, MatrixFactorizationModel


        # notice how similar the usage is to the standard ``CollieTrainer``
        model = MatrixFactorizationModel(train=train)
        trainer = CollieMinimalTrainer(model)
        trainer.fit(model)

    Model results should NOT be significantly different whether trained with ``CollieTrainer`` or
    ``CollieMinimalTrainer``.

    If there's an argument you would like to see added to ``CollieMinimalTrainer`` that is present
    in ``CollieTrainer`` used during productionalized model training, make an Issue or a PR in
    GitHub!

    Parameters
    ----------
    model: collie.model.BasePipeline
        Initialized Collie model
    max_epochs: int
        Stop training once this number of epochs is reached
    gpus: bool or int
        Whether to train on the GPU (``gpus == True`` or ``gpus > 0``) or the CPU
    logger: LightningLoggerBase
        Logger for experiment tracking. Set ``logger = None`` or ``logger = False`` to disable
        logging
    early_stopping_patience: int
        Number of epochs of patience to have without any improvement in loss before stopping
        training early. Validation epoch loss will be used if there is a validation DataLoader
        present, else training epoch loss will be used. Set ``early_stopping_patience = None`` or
        ``early_stopping_patience = False`` to disable early stopping
    log_every_n_steps: int
        How often to log within steps, if ``logger`` is enabled
    flush_logs_every_n_steps: int
        How often to flush logs to disk, if ``logger`` is enabled
    weights_summary: str
        Prints a summary of the weights when training begins
    terminate_on_nan: bool
        If set to ``True``, will terminate training (by raising a ``ValueError``) at the end of each
        training batch, if any of the parameters or the loss are NaN or +/- infinity
    benchmark: bool
        If set to ``True``, enables ``cudnn.benchmark``
    deterministic: bool
        If set to ``True``, enables ``cudnn.deterministic``
    progress_bar_refresh_rate: int
        How often to refresh progress bar (in steps), if ``verbosity > 0``
    verbosity: Union[bool, int]
        How verbose to be in training.

        * ``0`` disables all printouts, including ``weights_summary``

        * ``1`` prints ``weights_summary`` (if applicable) and epoch losses

        * ``2`` prints ``weights_summary`` (if applicable), epoch losses, and progress bars

    """
    def __init__(self,
                 model: BasePipeline,
                 max_epochs: int = 10,
                 gpus: Optional[Union[bool, int]] = None,
                 logger: Optional[LightningLoggerBase] = None,
                 early_stopping_patience: Optional[int] = 3,
                 log_every_n_steps: int = 50,
                 flush_logs_every_n_steps: int = 100,
                 weights_summary: Optional[str] = 'top',
                 terminate_on_nan: bool = False,
                 benchmark: bool = True,
                 deterministic: bool = True,
                 progress_bar_refresh_rate: Optional[int] = None,
                 verbosity: Union[bool, int] = True):
        # some light argument validation before saving as class-level attributes
        if gpus is None and torch.cuda.is_available():
            print('Detected GPU. Setting ``gpus`` to 1.')
            gpus = 1

        if logger is False:
            logger = None

        if early_stopping_patience is False:
            early_stopping_patience = None

        if verbosity is True:
            verbosity = 2
        elif verbosity is False:
            verbosity = 0

        self.max_epochs = max_epochs
        self.gpus = gpus
        self.benchmark = benchmark
        self.deterministic = deterministic
        self.logger = logger
        self.early_stopping_patience = early_stopping_patience
        self.log_every_n_steps = log_every_n_steps
        self.flush_logs_every_n_steps = flush_logs_every_n_steps
        self.weights_summary = weights_summary
        self.terminate_on_nan = terminate_on_nan
        self.progress_bar_refresh_rate = progress_bar_refresh_rate
        self.verbosity = verbosity

        self.best_epoch_loss = (0, sys.maxsize)
        self.train_steps = 0
        self.val_steps = 0
        self.num_epochs_completed = 0

        if self.gpus is None or self.gpus is False or self.gpus == 0:
            self.device = 'cpu'
        else:
            self.device = 'cuda'

        torch.backends.cudnn.benchmark = self.benchmark
        torch.backends.cudnn.deterministic = self.deterministic

    def fit(self, model: BasePipeline) -> None:
        """
        Runs the full optimization routine.

        Parameters
        ----------
        model: collie.model.BasePipeline
            Initialized Collie model

        """
        if (
            not hasattr(self, 'first_run_pre_training_setup_complete_')
            or not self.first_run_pre_training_setup_complete_
        ):
            self._pre_training_setup(model=model)
            self.first_run_pre_training_setup_complete_ = True

        self._initialize_optimizers_and_lr_schedulers(model=model)

        # set up top-level epoch progress bar
        epoch_iterator = range(self.num_epochs_completed + 1, self.max_epochs + 1)
        if self.verbosity >= 2:
            epoch_iterator = tqdm(epoch_iterator,
                                  position=0,
                                  unit='epoch',
                                  desc='',
                                  miniters=self.progress_bar_refresh_rate)

        for epoch in epoch_iterator:
            # run the training loop
            model.train()
            train_loss = self._train_loop_single_epoch(model, epoch)
            model.eval()

            epoch_summary = f'Epoch {epoch: >5}: train loss: {train_loss :<1.5f}'
            early_stop_loss = train_loss

            # save epoch loss metrics to the logger
            if self.logger is not None:
                self.logger.log_metrics(metrics={'train_loss_epoch': train_loss}, step=epoch)

            # run the validation loop logic, if we have the ``val_dataloader`` to do so
            if self.val_dataloader is not None:
                val_loss = self._val_loop_single_epoch(model)
                epoch_summary += f', val loss: {val_loss :<1.5f}'
                early_stop_loss = val_loss

                if self.logger is not None:
                    self.logger.log_metrics(metrics={'val_loss_epoch': val_loss}, step=epoch)

            # write out to disk only a single time at the end of the epoch
            if self.logger is not None:
                self.logger.save()

            if self.verbosity >= 1:
                print(epoch_summary)

            model.hparams.num_epochs_completed += 1
            self.num_epochs_completed += 1

            # early stopping logic
            if (
                self.early_stopping_patience is not None
                and early_stop_loss >= self.best_epoch_loss[1]
                and epoch >= (self.early_stopping_patience + self.best_epoch_loss[0])
            ):
                print(f'Epoch {epoch :>5}: Early stopping activated.')
                self._finalize_training()
                return

            # save best loss stats for future early stopping logic
            if early_stop_loss < self.best_epoch_loss[1]:
                self.best_epoch_loss = (epoch, early_stop_loss)

            # learning rate scheduler stepping, if applicable
            if self.lr_scheduler is not None:
                try:
                    # used for most learning rate schedulers
                    self.lr_scheduler.step()
                except TypeError:
                    # used for ``ReduceLROnPlateau``
                    self.lr_scheduler.step(early_stop_loss)

        # run final logging things when training is complete before returning
        self._finalize_training()

    def _pre_training_setup(self, model: BasePipeline) -> None:
        """Set up DataLoaders, optimizers, learning rate schedulers, etc. before training starts."""
        self.train_dataloader = model.train_dataloader()
        self.val_dataloader = model.val_dataloader()

        if self.verbosity != 0 and self.weights_summary is not None:
            print(ModelSummary(model, mode=self.weights_summary))

        # log model hyperparameters, if applicable
        if self.logger is not None:
            self.logger.log_hyperparams(model.hparams)
            self.logger.save()

        # move the model over to the device
        model.to(self.device)

    def _initialize_optimizers_and_lr_schedulers(self, model: BasePipeline) -> None:
        self.lr_scheduler = None
        configure_optimizers_return_value = model.configure_optimizers()
        if isinstance(configure_optimizers_return_value, tuple):
            # we have a list of optimizers and a list of lr_schedulers dictionaries
            optimizers, lr_schedulers = configure_optimizers_return_value
            self.optimizer = MultiOptimizer(optimizers)
            self.lr_scheduler = MultiLRScheduler(lr_schedulers)
        elif isinstance(configure_optimizers_return_value, list):
            # we have a list of optimizers
            self.optimizer = MultiOptimizer(configure_optimizers_return_value)
        elif isinstance(configure_optimizers_return_value, torch.optim.Optimizer):
            # we have a single optimizer
            self.optimizer = MultiOptimizer([configure_optimizers_return_value])
        else:
            # we have something we've never seen before
            raise ValueError('Unexpected output from ``model.configure_optimizers()``!')

    def _train_loop_single_epoch(self, model: torch.nn.Module, epoch: int) -> float:
        """Training loop for a single epoch, where gradients are optimized for."""
        total_loss = 0

        train_dataloader_iterator = enumerate(self.train_dataloader)
        if self.verbosity >= 2:
            train_dataloader_iterator = tqdm(train_dataloader_iterator,
                                             total=len(self.train_dataloader),
                                             unit='step',
                                             desc=f'({epoch :^5})',
                                             leave=False,
                                             miniters=self.progress_bar_refresh_rate)

        for batch_idx, batch in train_dataloader_iterator:
            self.optimizer.zero_grad()

            batch = self._move_batch_to_device(batch)
            loss = model.calculate_loss(batch)
            loss.backward()

            for optimizer_idx, optimizer in enumerate(self.optimizer.optimizers):
                model.optimizer_step(epoch=epoch,
                                     batch_idx=batch_idx,
                                     optimizer=optimizer,
                                     optimizer_idx=optimizer_idx,
                                     optimizer_closure=None)

            self.train_steps += 1

            if self.terminate_on_nan and not torch.isfinite(loss).all():
                raise ValueError(f'Loss is {loss}, stopping training early!')

            detached_loss = loss.detach()
            total_loss += detached_loss

            if self.verbosity >= 2:
                train_dataloader_iterator.set_postfix(train_loss=detached_loss.item())

            self._log_step(name='train',
                           steps=self.train_steps,
                           total_loss=total_loss,
                           batch_idx=batch_idx)

        return (total_loss / len(self.train_dataloader)).item()

    def _val_loop_single_epoch(self, model: torch.nn.Module) -> float:
        """Validation loop for a single epoch, where gradients are NOT optimized for."""
        total_loss = 0

        for batch_idx, batch in enumerate(self.val_dataloader):
            batch = self._move_batch_to_device(batch)
            loss = model.calculate_loss(batch)

            self.val_steps += 1

            total_loss += loss.detach()

            self._log_step(name='val',
                           steps=self.val_steps,
                           total_loss=total_loss,
                           batch_idx=batch_idx)

        return (total_loss / len(self.val_dataloader)).item()

    def _move_batch_to_device(
        self, batch: Tuple[Tuple[torch.tensor, torch.tensor], torch.tensor],
    ) -> Tuple[Tuple[torch.tensor, torch.tensor], torch.tensor]:
        """Move a batch of data to the proper device."""
        # TODO: does this actually speed anything up?
        try:
            # assume we have implicit data
            ((users, pos_items), neg_items) = batch

            users = users.to(self.device)
            pos_items = pos_items.to(self.device)
            neg_items = neg_items.to(self.device)

            return ((users, pos_items), neg_items)
        except (AttributeError, ValueError):
            try:
                # now assume we have explicit data
                users, pos_items, ratings = batch

                users = users.to(self.device)
                pos_items = pos_items.to(self.device)
                ratings = ratings.to(self.device)

                return users, pos_items, ratings
            except (AttributeError, ValueError):
                # we have an unexpected data format, fallback to PyTorch Lightning
                return move_data_to_device(batch, self.device)

    def _log_step(self, name: str, steps: int, total_loss: torch.tensor, batch_idx: int) -> None:
        """Check if we should and, if so, log step-loss metrics to our logger."""
        if self.logger is not None:
            if steps % self.log_every_n_steps == 0:
                batch_loss = (total_loss / (batch_idx + 1)).item()
                self.logger.log_metrics(metrics={f'{name}_loss_step': batch_loss}, step=steps)
            if steps % self.flush_logs_every_n_steps == 0:
                self.logger.save()

    def _finalize_training(self) -> None:
        """Finalize logging results before returning."""
        if self.logger is not None:
            self.logger.save()
            self.logger.finalize(status='FINISHED')



from functools import partial
from typing import Callable, Dict, Optional, Union

import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding
from collie.utils import get_init_arguments, merge_docstrings, trunc_normal


class NeuralCollaborativeFiltering(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for a neural matrix factorization model.

    ``NeuralCollaborativeFiltering`` models combine a collaborative filtering and multilayer
    perceptron network in a single, unified model. The model consists of two sections: the first
    is a simple matrix factorization that calculates a score by multiplying together user and item
    embeddings (lookups through an embedding table); the second is a MLP network that feeds
    embeddings from a second set of embedding tables (one for user, one for item). Both output
    vectors are combined and sent through a final MLP layer before returning a single recommendation
    score.

    The implementation here is meant to mimic its original implementation as specified here:
    https://arxiv.org/pdf/1708.05031.pdf [2]_

    All ``NeuralCollaborativeFiltering`` instances are subclasses of the ``LightningModule`` class
    provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, NeuralCollaborativeFiltering


        model = NeuralCollaborativeFiltering(train=train)
        trainer = CollieTrainer(model)
        trainer.fit(model)
        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(filename='model.pth')
        new_model = NeuralCollaborativeFiltering(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Parameters
    ----------
    embedding_dim: int
        Number of latent factors to use for the matrix factorization embedding table. For the MLP
        embedding table, the dimensionality will be calculated with the formula
        ``embedding_dim * (2 ** (num_layers - 1))``
    num_layers: int
        Number of MLP layers to apply. Each MLP layer will have its input dimension calculated with
        the formula ``embedding_dim * (2 ** (``num_layers`` - ``current_layer_number``))``
    final_layer: str or function
        Final layer activation function. Available string options include:

        * 'sigmoid'

        * 'relu'

        * 'leaky_relu'

    dropout_p: float
        Probability of dropout on the MLP layers
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    References
    ----------
    .. [2] Xiangnan et al. "Neural Collaborative Filtering." Neural Collaborative Filtering |
        Proceedings of the 26th International Conference on World Wide Web, 1 Apr. 2017,
        dl.acm.org/doi/10.1145/3038912.3052569.

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 embedding_dim: int = 8,
                 num_layers: int = 3,
                 final_layer: Optional[Union[str, Callable]] = None,
                 dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 # y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        super().__init__(**get_init_arguments())

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after ``prepare_data``.

        """
        self.user_embeddings_cf = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                                  embedding_dim=self.hparams.embedding_dim)
        self.item_embeddings_cf = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                                  embedding_dim=self.hparams.embedding_dim)

        mlp_embedding_dim = self.hparams.embedding_dim * (2 ** (self.hparams.num_layers - 1))
        self.user_embeddings_mlp = ScaledEmbedding(
            num_embeddings=self.hparams.num_users,
            embedding_dim=mlp_embedding_dim,
        )
        self.item_embeddings_mlp = ScaledEmbedding(
            num_embeddings=self.hparams.num_items,
            embedding_dim=mlp_embedding_dim,
        )

        mlp_modules = []
        for i in range(self.hparams.num_layers):
            input_size = self.hparams.embedding_dim * (2 ** (self.hparams.num_layers - i))
            mlp_modules.append(nn.Dropout(p=self.hparams.dropout_p))
            mlp_modules.append(nn.Linear(input_size, input_size//2))
            mlp_modules.append(nn.ReLU())
        self.mlp_layers = nn.Sequential(*mlp_modules)

        self.predict_layer = nn.Linear(self.hparams.embedding_dim * 2, 1)

        for m in self.mlp_layers:
            if isinstance(m, nn.Linear):
                # initialization taken from the official repo:
                # https://github.com/hexiangnan/neural_collaborative_filtering/blob/master/NeuMF.py#L63  # noqa: E501
                trunc_normal(m.weight.data, std=0.01)

        nn.init.kaiming_uniform_(self.predict_layer.weight, nonlinearity='relu')

        for m in self.modules():
            if isinstance(m, nn.Linear) and m.bias is not None:
                m.bias.data.zero_()

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embedding_cf = self.user_embeddings_cf(users)
        item_embedding_cf = self.item_embeddings_cf(items)
        output_cf = user_embedding_cf * item_embedding_cf

        user_embedding_mlp = self.user_embeddings_mlp(users)
        item_embedding_mlp = self.item_embeddings_mlp(items)
        interaction = torch.cat((user_embedding_mlp, item_embedding_mlp), -1)
        output_mlp = self.mlp_layers(interaction)

        concat = torch.cat((output_cf, output_mlp), -1)

        prediction = self.predict_layer(concat)

        if callable(self.hparams.final_layer):
            prediction = self.hparams.final_layer(prediction)
        elif self.hparams.final_layer == 'sigmoid':
            prediction = torch.sigmoid(prediction)
        elif self.hparams.final_layer == 'relu':
            prediction = F.relu(prediction)
        elif self.hparams.final_layer == 'leaky_relu':
            prediction = F.leaky_relu(prediction)
        elif self.hparams.final_layer is not None:
            raise ValueError(f'{self.hparams.final_layer} not valid final layer value!')

        return prediction.view(-1)

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings, which are the concatenated CF and MLP item embeddings, on device."""
        items = torch.arange(self.hparams.num_items, device=self.device)

        return torch.cat((
            self.item_embeddings_cf(items),
            self.item_embeddings_mlp(items),
        ), axis=1).detach()



import os
import tempfile

import pytest
import torch

from collie.interactions import (ApproximateNegativeSamplingInteractionsDataLoader,
                                 HDF5InteractionsDataLoader,
                                 InteractionsDataLoader)
from collie.model import (ColdStartModel,
                          CollaborativeMetricLearningModel,
                          CollieMinimalTrainer,
                          CollieTrainer,
                          DeepFM,
                          HybridModel,
                          HybridPretrainedModel,
                          MatrixFactorizationModel,
                          MLPMatrixFactorizationModel,
                          NeuralCollaborativeFiltering,
                          NonlinearMatrixFactorizationModel)
from collie.utils import pandas_df_to_hdf5


@pytest.fixture(scope='session')
def implicit_model(train_val_implicit_data, gpu_count):
    train, val = train_val_implicit_data
    model = MatrixFactorizationModel(train=train,
                                     val=val,
                                     embedding_dim=10,
                                     lr=1e-1)
    model_trainer = CollieTrainer(model=model,
                                  gpus=gpu_count,
                                  max_epochs=10,
                                  deterministic=True,
                                  logger=False,
                                  checkpoint_callback=False)
    model_trainer.fit(model)
    model.eval()

    return model


@pytest.fixture(scope='session')
def implicit_model_no_lightning(train_val_implicit_data, gpu_count):
    train, val = train_val_implicit_data
    model = MatrixFactorizationModel(train=train,
                                     val=val,
                                     embedding_dim=10,
                                     lr=1e-1)
    model_trainer = CollieMinimalTrainer(model=model,
                                         gpus=gpu_count,
                                         max_epochs=10,
                                         deterministic=True,
                                         logger=False,
                                         early_stopping_patience=False)
    model_trainer.fit(model)
    model.freeze()

    return model


@pytest.fixture(scope='session')
def explicit_model(train_val_explicit_data, gpu_count):
    train, val = train_val_explicit_data
    model = MatrixFactorizationModel(train=train,
                                     val=val,
                                     embedding_dim=10,
                                     lr=1e-2,
                                     loss='mse',
                                     y_range=[1, 5])
    model_trainer = CollieTrainer(model=model,
                                  gpus=gpu_count,
                                  max_epochs=10,
                                  deterministic=True,
                                  logger=False,
                                  checkpoint_callback=False)
    model_trainer.fit(model)
    model.freeze()

    return model


@pytest.fixture(scope='session')
def explicit_model_no_lightning(train_val_explicit_data, gpu_count):
    train, val = train_val_explicit_data
    model = MatrixFactorizationModel(train=train,
                                     val=val,
                                     embedding_dim=10,
                                     lr=1e-2,
                                     loss='mse',
                                     y_range=[1, 5])
    model_trainer = CollieMinimalTrainer(model=model,
                                         gpus=gpu_count,
                                         max_epochs=10,
                                         deterministic=True,
                                         logger=False,
                                         early_stopping_patience=False)
    model_trainer.fit(model)
    model.freeze()

    return model


@pytest.fixture(scope='session')
def untrained_implicit_model(train_val_implicit_data):
    train, val = train_val_implicit_data
    model = MatrixFactorizationModel(train=train, val=val)

    return model


@pytest.fixture(scope='session')
def untrained_implicit_model_no_val_data(train_val_implicit_data):
    train, _ = train_val_implicit_data
    model = MatrixFactorizationModel(train=train, val=None)

    return model


@pytest.fixture(params=['mf_hdf5',
                        'mf_with_y_range',
                        'collaborative_metric_learning',
                        'sparse_collaborative_metric_learning',
                        'mlp_mf',
                        'mlp_mf_with_y_range',
                        'sparse_mf',
                        'mf_no_val',
                        'mf_non_approximate',
                        'mf_approximate',
                        'nonlinear_mf',
                        'nonlinear_mf_with_y_range',
                        'neucf',
                        'neucf_sigmoid',
                        'neucf_relu',
                        'neucf_leaky_rulu',
                        'neucf_custom',
                        'deep_fm',
                        'deep_fm_sigmoid',
                        'deep_fm_relu',
                        'deep_fm_leaky_rulu',
                        'deep_fm_custom',
                        'hybrid_pretrained',
                        'hybrid_pretrained_metadata_layers',
                        'hybrid_mf',
                        'hybrid_mf_metadata_layers',
                        'hybrid_mf_metadata_layers_metadata_for_loss',
                        'hybrid_mf_metadata_layers_no_bias_optimizer',
                        'cold_start',
                        'cold_start_metadata_for_loss',
                        'cold_start_no_val',
                        'cold_start_no_lr_scheduler'])
def models_trained_for_one_step(request,
                                train_val_implicit_sample_data,
                                movielens_metadata_df,
                                movielens_implicit_df,
                                train_val_implicit_pandas_data,
                                gpu_count):
    train, val = train_val_implicit_sample_data

    genres = (
        torch.tensor(movielens_metadata_df[
            [c for c in movielens_metadata_df.columns if 'genre' in c]
        ].values)
        .topk(1)
        .indices
        .view(-1)
    )

    number_of_stages = None

    if request.param == 'mf_hdf5':
        # create, fit, and return the model all at once so we can close the HDF5 file
        train_pandas_df, val_pandas_df = train_val_implicit_pandas_data

        with tempfile.TemporaryDirectory() as temp_dir:
            pandas_df_to_hdf5(df=train_pandas_df,
                              out_path=os.path.join(temp_dir, 'train.h5'),
                              key='interactions')
            pandas_df_to_hdf5(df=val_pandas_df,
                              out_path=os.path.join(temp_dir, 'val.h5'),
                              key='interactions')

            train_loader = HDF5InteractionsDataLoader(hdf5_path=os.path.join(temp_dir, 'train.h5'),
                                                      user_col='user_id',
                                                      item_col='item_id',
                                                      num_users=train.num_users,
                                                      num_items=train.num_items,
                                                      batch_size=1024,
                                                      shuffle=True)
            val_loader = HDF5InteractionsDataLoader(hdf5_path=os.path.join(temp_dir, 'val.h5'),
                                                    user_col='user_id',
                                                    item_col='item_id',
                                                    num_users=val.num_users,
                                                    num_items=val.num_items,
                                                    batch_size=1024,
                                                    shuffle=False)

            model = MatrixFactorizationModel(train=train_loader,
                                             val=val_loader,
                                             embedding_dim=15,
                                             dropout_p=0.1,
                                             lr=1e-1,
                                             bias_lr=1e-2,
                                             optimizer='adam',
                                             bias_optimizer='sgd',
                                             weight_decay=1e-7,
                                             loss='bpr',
                                             sparse=False)

            model_trainer = CollieTrainer(model=model,
                                          gpus=gpu_count,
                                          max_steps=1,
                                          deterministic=True,
                                          logger=False,
                                          checkpoint_callback=False)

            model_trainer.fit(model)
            model.eval()

            return model

    elif request.param == 'sparse_mf':
        model = MatrixFactorizationModel(train=train,
                                         val=val,
                                         embedding_dim=15,
                                         dropout_p=0.1,
                                         lr=1e-1,
                                         bias_lr=1e-2,
                                         optimizer='sparse_adam',
                                         bias_optimizer='sgd',
                                         weight_decay=0,
                                         loss='hinge',
                                         sparse=True)
    elif request.param == 'mf_no_val':
        model = MatrixFactorizationModel(train=train, val=None)
    elif request.param == 'mf_non_approximate' or request.param == 'mf_approximate':
        if request.param == 'mf_non_approximate':
            train_loader = InteractionsDataLoader(interactions=train, batch_size=1024, shuffle=True)
            val_loader = InteractionsDataLoader(interactions=val, batch_size=1024, shuffle=False)
        else:
            train_loader = ApproximateNegativeSamplingInteractionsDataLoader(interactions=train,
                                                                             batch_size=1024,
                                                                             shuffle=True)
            val_loader = ApproximateNegativeSamplingInteractionsDataLoader(interactions=val,
                                                                           batch_size=1024,
                                                                           shuffle=False)

        model = MatrixFactorizationModel(train=train_loader,
                                         val=val_loader,
                                         embedding_dim=15,
                                         dropout_p=0.1,
                                         lr=1e-1,
                                         bias_lr=1e-2,
                                         optimizer='adam',
                                         bias_optimizer='sgd',
                                         weight_decay=1e-7,
                                         loss='bpr',
                                         sparse=False)
    elif request.param == 'mf_with_y_range':
        model = MatrixFactorizationModel(train=train,
                                         val=val,
                                         y_range=(0, 4))
    elif request.param == 'collaborative_metric_learning':
        model = CollaborativeMetricLearningModel(train=train,
                                                 val=val,
                                                 embedding_dim=15,
                                                 lr=1e-1,
                                                 optimizer='adam',
                                                 loss='hinge')
    elif request.param == 'sparse_collaborative_metric_learning':
        model = CollaborativeMetricLearningModel(train=train,
                                                 val=val,
                                                 embedding_dim=15,
                                                 lr=1e-1,
                                                 optimizer='sparse_adam',
                                                 loss='hinge',
                                                 sparse=True)
    elif request.param == 'mlp_mf':
        model = MLPMatrixFactorizationModel(train=train,
                                            val=val,
                                            embedding_dim=15,
                                            num_layers=3,
                                            dropout_p=0.1,
                                            lr=1e-1,
                                            bias_lr=1e-2,
                                            optimizer='adam',
                                            bias_optimizer='sgd',
                                            weight_decay=1e-7,
                                            loss='hinge')
    elif request.param == 'mlp_mf_with_y_range':
        model = MLPMatrixFactorizationModel(train=train,
                                            val=val,
                                            y_range=(0, 2))
    elif request.param == 'nonlinear_mf':
        model = NonlinearMatrixFactorizationModel(train=train,
                                                  val=val,
                                                  user_embedding_dim=15,
                                                  item_embedding_dim=15,
                                                  user_dense_layers_dims=[15, 10],
                                                  item_dense_layers_dims=[15, 10],
                                                  embedding_dropout_p=0.05,
                                                  dense_dropout_p=0.1,
                                                  lr=1e-2,
                                                  bias_lr=1e-1,
                                                  optimizer='adam',
                                                  bias_optimizer='sgd',
                                                  weight_decay=1e-7,
                                                  loss='bpr')
    elif request.param == 'nonlinear_mf_with_y_range':
        model = NonlinearMatrixFactorizationModel(train=train,
                                                  val=val,
                                                  y_range=(0, 4))
    elif request.param == 'neucf':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             embedding_dim=10,
                                             num_layers=1,
                                             dropout_p=0.1,
                                             lr=1e-4,
                                             weight_decay=0.,
                                             optimizer='adam',
                                             loss='adaptive')
    elif request.param == 'neucf_sigmoid':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             lr=1e-4,
                                             final_layer='sigmoid')
    elif request.param == 'neucf_relu':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             lr=1e-4,
                                             final_layer='relu')
    elif request.param == 'neucf_leaky_rulu':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             lr=1e-4,
                                             final_layer='leaky_relu')
    elif request.param == 'neucf_custom':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             lr=1e-4,
                                             final_layer=torch.tanh)
    elif request.param == 'deep_fm':
        model = DeepFM(train=train,
                       val=val,
                       embedding_dim=10,
                       num_layers=1,
                       dropout_p=0.1,
                       lr=1e-3,
                       weight_decay=0.,
                       optimizer='adam',
                       bias_optimizer=None,
                       loss='hinge')
    elif request.param == 'deep_fm_sigmoid':
        model = DeepFM(train=train,
                       val=val,
                       final_layer='sigmoid')
    elif request.param == 'deep_fm_relu':
        model = DeepFM(train=train,
                       val=val,
                       final_layer='relu')
    elif request.param == 'deep_fm_leaky_rulu':
        model = DeepFM(train=train,
                       val=val,
                       final_layer='leaky_relu')
    elif request.param == 'deep_fm_custom':
        model = DeepFM(train=train,
                       val=val,
                       final_layer=torch.tanh)
    elif (
        request.param == 'hybrid_pretrained' or request.param == 'hybrid_pretrained_metadata_layers'
    ):
        implicit_model = MatrixFactorizationModel(train=train,
                                                  val=val,
                                                  embedding_dim=10,
                                                  lr=1e-1,
                                                  optimizer='adam')
        implicit_model_trainer = CollieTrainer(model=implicit_model,
                                               gpus=gpu_count,
                                               max_steps=1,
                                               deterministic=True,
                                               logger=False,
                                               checkpoint_callback=False)
        implicit_model_trainer.fit(implicit_model)
        implicit_model.eval()

        metadata_layers_dims = None
        if request.param == 'hybrid_pretrained_metadata_layers':
            metadata_layers_dims = [16, 12]

        model_frozen = HybridPretrainedModel(train=train,
                                             val=val,
                                             item_metadata=movielens_metadata_df,
                                             trained_model=implicit_model,
                                             metadata_layers_dims=metadata_layers_dims,
                                             freeze_embeddings=True,
                                             dropout_p=0.15,
                                             loss='warp',
                                             lr=.01,
                                             optimizer=torch.optim.Adam,
                                             weight_decay=0.0)
        model_frozen_trainer = CollieTrainer(model=model_frozen,
                                             gpus=gpu_count,
                                             max_steps=1,
                                             deterministic=True,
                                             logger=False,
                                             checkpoint_callback=False)
        model_frozen_trainer.fit(model_frozen)

        model = HybridPretrainedModel(train=train,
                                      val=val,
                                      item_metadata=movielens_metadata_df,
                                      trained_model=implicit_model,
                                      metadata_layers_dims=metadata_layers_dims,
                                      freeze_embeddings=False,
                                      dropout_p=0.15,
                                      loss='bpr',
                                      lr=1e-4,
                                      optimizer=torch.optim.Adam,
                                      metadata_for_loss={'genre': genres},
                                      metadata_for_loss_weights={'genre': .4},
                                      weight_decay=0.0)
        model.load_from_hybrid_model(model_frozen)
    elif (
        request.param == 'hybrid_mf'
        or request.param == 'hybrid_mf_metadata_layers'
        or request.param == 'hybrid_mf_metadata_layers_metadata_for_loss'
        or request.param == 'hybrid_mf_metadata_layers_no_bias_optimizer'
    ):
        number_of_stages = 3

        metadata_layers_dims = None
        if request.param == 'hybrid_mf_metadata_layers':
            metadata_layers_dims = [16, 12]

        additional_kwargs = {}
        if request.param == 'hybrid_mf_metadata_layers_metadata_for_loss':
            additional_kwargs = {
                'metadata_for_loss': {'genre': genres},
                'metadata_for_loss_weights': {'genre': .4},
            }
        elif request.param == 'hybrid_mf_metadata_layers_no_bias_optimizer':
            additional_kwargs = {
                'bias_optimizer': None
            }

        model = HybridModel(train=train,
                            val=val,
                            item_metadata=movielens_metadata_df,
                            embedding_dim=10,
                            metadata_layers_dims=metadata_layers_dims,
                            lr=1e-1,
                            optimizer='adam',
                            **additional_kwargs)
    elif (
        request.param == 'cold_start'
        or request.param == 'cold_start_metadata_for_loss'
        or request.param == 'cold_start_no_val'
        or request.param == 'cold_start_no_lr_scheduler'
    ):
        number_of_stages = 2

        additional_kwargs = {}
        if request.param == 'cold_start_metadata_for_loss':
            additional_kwargs = {
                'metadata_for_loss': {'genre': genres},
                'metadata_for_loss_weights': {'genre': .4},
            }
        elif request.param == 'cold_start_no_lr_scheduler':
            additional_kwargs = {
                'lr_scheduler_func': None
            }

        model = ColdStartModel(train=train,
                               val=val if request.param != 'cold_start_no_val' else None,
                               item_buckets=genres[:train.num_items],
                               embedding_dim=10,
                               item_buckets_stage_lr=1e-2,
                               no_buckets_stage_lr=1e-2,
                               item_buckets_stage_optimizer='sgd',
                               **additional_kwargs)

    model_trainer = CollieTrainer(model=model,
                                  gpus=gpu_count,
                                  max_steps=1,
                                  deterministic=True,
                                  logger=False,
                                  checkpoint_callback=False)

    if request.param == 'mf_no_val':
        with pytest.warns(UserWarning):
            model_trainer.fit(model)
    else:
        if number_of_stages is None:
            model_trainer.fit(model)
        else:
            for idx in range(number_of_stages):
                model_trainer.fit(model)

                if idx < (number_of_stages - 1):
                    model_trainer.max_steps += 1
                    model.advance_stage()

    model.eval()

    return model


@pytest.fixture(params=['mf',
                        'sparse_mf',
                        'nonlinear_mf',
                        'neucf',
                        'hybrid_pretrained',
                        'hybrid_mf',
                        'cold_start'])
def explicit_models_trained_for_one_step(request,
                                         train_val_explicit_sample_data,
                                         movielens_metadata_df,
                                         gpu_count):
    train, val = train_val_explicit_sample_data

    genres = (
        torch.tensor(movielens_metadata_df[
            [c for c in movielens_metadata_df.columns if 'genre' in c]
        ].values)
        .topk(1)
        .indices
        .view(-1)
    )

    number_of_stages = None

    if request.param == 'mf':
        model = MatrixFactorizationModel(train=train, val=val, loss='mse')
    if request.param == 'sparse_mf':
        model = MatrixFactorizationModel(train=train,
                                         val=val,
                                         embedding_dim=15,
                                         dropout_p=0.1,
                                         lr=1e-1,
                                         bias_lr=1e-2,
                                         optimizer='sparse_adam',
                                         bias_optimizer='sgd',
                                         weight_decay=0,
                                         loss='mse',
                                         sparse=True)
    elif request.param == 'nonlinear_mf':
        model = NonlinearMatrixFactorizationModel(train=train,
                                                  val=val,
                                                  user_embedding_dim=15,
                                                  item_embedding_dim=15,
                                                  user_dense_layers_dims=[15, 10],
                                                  item_dense_layers_dims=[15, 10],
                                                  embedding_dropout_p=0.05,
                                                  dense_dropout_p=0.1,
                                                  lr=1e-1,
                                                  bias_lr=1e-2,
                                                  optimizer='adam',
                                                  bias_optimizer='sgd',
                                                  weight_decay=1e-7,
                                                  loss='mse')
    elif request.param == 'neucf':
        model = NeuralCollaborativeFiltering(train=train,
                                             val=val,
                                             embedding_dim=10,
                                             num_layers=1,
                                             dropout_p=0.1,
                                             lr=1e-4,
                                             weight_decay=0.,
                                             optimizer='adam',
                                             loss='mae')
    elif request.param == 'hybrid_pretrained':
        implicit_model = MatrixFactorizationModel(train=train,
                                                  val=val,
                                                  embedding_dim=10,
                                                  lr=1e-1,
                                                  optimizer='adam',
                                                  loss='mse')
        implicit_model_trainer = CollieTrainer(model=implicit_model,
                                               gpus=gpu_count,
                                               max_steps=1,
                                               deterministic=True,
                                               logger=False,
                                               checkpoint_callback=False)
        implicit_model_trainer.fit(implicit_model)
        implicit_model.freeze()

        genres = (
            torch.tensor(movielens_metadata_df[
                [c for c in movielens_metadata_df.columns if 'genre' in c]
            ].values)
            .topk(1)
            .indices
            .view(-1)
        )

        model_frozen = HybridPretrainedModel(train=train,
                                             val=val,
                                             item_metadata=movielens_metadata_df,
                                             trained_model=implicit_model,
                                             metadata_layers_dims=None,
                                             freeze_embeddings=True,
                                             dropout_p=0.15,
                                             loss='mae',
                                             lr=.01,
                                             optimizer=torch.optim.Adam,
                                             metadata_for_loss={'genre': genres},
                                             metadata_for_loss_weights={'genre': .4},
                                             weight_decay=0.0)
        model_frozen_trainer = CollieTrainer(model=model_frozen,
                                             gpus=gpu_count,
                                             max_steps=1,
                                             deterministic=True,
                                             logger=False,
                                             checkpoint_callback=False)
        model_frozen_trainer.fit(model_frozen)

        model = HybridPretrainedModel(train=train,
                                      val=val,
                                      item_metadata=movielens_metadata_df,
                                      trained_model=implicit_model,
                                      metadata_layers_dims=None,
                                      freeze_embeddings=False,
                                      dropout_p=0.15,
                                      loss='mse',
                                      lr=1e-4,
                                      optimizer=torch.optim.Adam,
                                      metadata_for_loss={'genre': genres},
                                      metadata_for_loss_weights={'genre': .4},
                                      weight_decay=0.0)
        model.load_from_hybrid_model(model_frozen)
    elif request.param == 'hybrid_mf':
        number_of_stages = 3

        model = HybridModel(train=train,
                            val=val,
                            item_metadata=movielens_metadata_df,
                            embedding_dim=10,
                            metadata_layers_dims=[16, 12],
                            lr=1e-1,
                            loss='mae',
                            optimizer='adam')
    elif request.param == 'cold_start':
        number_of_stages = 2

        model = ColdStartModel(train=train,
                               val=val,
                               item_buckets=genres[:train.num_items],
                               embedding_dim=10,
                               item_buckets_stage_lr=1e-2,
                               no_buckets_stage_lr=1e-2,
                               loss='mse',
                               item_buckets_stage_optimizer='sgd')

    model_trainer = CollieTrainer(model=model,
                                  gpus=gpu_count,
                                  max_steps=1,
                                  deterministic=True,
                                  logger=False,
                                  checkpoint_callback=False)

    if request.param == 'mf_no_val':
        with pytest.warns(UserWarning):
            model_trainer.fit(model)
    else:
        if number_of_stages is None:
            model_trainer.fit(model)
        else:
            for idx in range(number_of_stages):
                model_trainer.fit(model)

                if idx < (number_of_stages - 1):
                    model_trainer.max_steps += 1
                    model.advance_stage()

    model.freeze()

    return model



from abc import ABCMeta, abstractmethod
import collections
import random
import textwrap
from typing import Any, Iterable, List, Optional, Tuple, Union
import warnings

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix, dok_matrix
import torch
from tqdm.auto import tqdm

import collie


class BaseInteractions(torch.utils.data.Dataset, metaclass=ABCMeta):
    """
    PyTorch ``Dataset`` for implicit user-item interactions data.

    If ``mat`` is provided, the ``Interactions`` instance will act as a wrapper for a sparse matrix
    in COOrdinate format, typically looking like:

    * Users comprising the rows

    * Items comprising the columns

    * Ratings given by that user for that item comprising the elements of the matrix

    ``Interactions`` can be instantiated instead by passing in single arrays with corresponding
    user_ids, item_ids, and ratings (by default, set to 1 for implicit recommenders) values with
    the same functionality as a matrix. Note that with this approach, the number of users and items
    will be the maximum values in those two columns, respectively, and it is expected that all
    integers between 0 and the maximum ID should appear somewhere in the data.

    By default, exact negative sampling will be used during each ``__getitem__`` call. To use
    approximate negative sampling, set ``max_number_of_samples_to_consider = 0``. This will avoid
    building a positive item lookup dictionary during initialization.

    Parameters
    ----------
    mat: scipy.sparse.coo_matrix or numpy.array, 2-dimensional
        Interactions matrix, which, if provided, will be used instead of ``users``, ``items``, and
        ``ratings`` arguments
    users: Iterable[int], 1-d
        Array of user IDs, starting at 0
    items: Iterable[int], 1-d
        Array of corresponding item IDs to ``users``, starting at 0
    ratings: Iterable[int], 1-d
        Array of corresponding ratings to both ``users`` and ``items``. If ``None``, will default to
        each user in ``user`` interacting with an item with a rating value of 1
    num_negative_samples: int
        Number of negative samples to return with each ``__getitem__`` call
    allow_missing_ids: bool
        If ``False``, will check that both ``users`` and ``items`` contain each integer from 0 to
        the maximum value in the array. This check only applies when initializing an
        ``Interactions`` instance using 1-dimensional arrays ``users`` and ``items``
    remove_duplicate_user_item_pairs: bool
        Will check for and remove any duplicate user, item ID pairs from the ``Interactions`` matrix
        during initialization. Note that this will create a second sparse matrix held in memory
        to efficiently check, which could cause memory concerns for larger data. If you are sure
        that there are no duplicated, user, item ID pairs, set to ``False``
    num_users: int
        Number of users in the dataset. If ``num_users == 'infer'``, this will be set to the
        ``mat.shape[0]`` or ``max(users) + 1``, depending on the input
    num_items: int
        Number of items in the dataset. If ``num_items == 'infer'``, this will be set to the
        ``mat.shape[1]`` or ``max(items) + 1``, depending on the input
    check_num_negative_samples_is_valid: bool
        Check that ``num_negative_samples`` is less than the maximum number of items a user has
        interacted with. If it is not, then for all users who have fewer than
        ``num_negative_samples`` items not interacted with, a random sample including positive items
        will be returned as negative
    max_number_of_samples_to_consider: int
        Number of samples to try for a given user before returning an approximate negative sample.
        This should be greater than ``num_negative_samples``. If set to ``0``, approximate negative
        sampling will be used by default in ``__getitem__`` and a positive item lookup dictionary
        will NOT be built

    """
    def __init__(self,
                 mat: Optional[Union[coo_matrix, np.array]] = None,
                 users: Optional[Iterable[int]] = None,
                 items: Optional[Iterable[int]] = None,
                 ratings: Optional[Iterable[int]] = None,
                 allow_missing_ids: bool = False,
                 remove_duplicate_user_item_pairs: bool = True,
                 num_users: int = 'infer',
                 num_items: int = 'infer'):
        if mat is None:
            assert users is not None and items is not None, (
                'Either 1) ``mat`` or 2) both ``users`` or ``items`` must be non-null!'
            )

            if len(users) != len(items):
                raise ValueError('Lengths of ``users`` and ``items`` must be equal.')

            num_users = collie.utils._infer_num_if_needed_for_1d_array(num_users, users)
            num_items = collie.utils._infer_num_if_needed_for_1d_array(num_items, items)

            if allow_missing_ids is False:
                _check_array_contains_all_integers(array=users,
                                                   array_max_value=num_users,
                                                   array_name='users')
                _check_array_contains_all_integers(array=items,
                                                   array_max_value=num_items,
                                                   array_name='items')

            if ratings is not None:
                if len(users) != len(ratings):
                    raise ValueError(
                        'Length of ``ratings`` must be equal to lengths of ``users``, ``items``.'
                    )

            mat = collie.utils._create_sparse_ratings_matrix_helper(users=users,
                                                                    items=items,
                                                                    ratings=ratings,
                                                                    num_users=num_users,
                                                                    num_items=num_items)
        else:
            mat = coo_matrix(mat)

            if num_users == 'infer':
                num_users = mat.shape[0]
            if num_items == 'infer':
                num_items = mat.shape[1]

            if allow_missing_ids is False:
                _check_array_contains_all_integers(array=mat.row,
                                                   array_max_value=num_users,
                                                   array_name='mat.shape[0]')
                _check_array_contains_all_integers(array=mat.col,
                                                   array_max_value=num_items,
                                                   array_name='mat.shape[1]')

        if remove_duplicate_user_item_pairs:
            print('Checking for and removing duplicate user, item ID pairs...')

            # remove duplicate entires in the COO matrix
            dok_mat = dok_matrix((mat.shape), dtype=mat.dtype)
            dok_mat._update(zip(zip(mat.row, mat.col), mat.data))
            mat = dok_mat.tocoo()

            # trigger garbage collection early
            del dok_mat

        self.mat = mat
        self.allow_missing_ids = allow_missing_ids
        self.remove_duplicate_user_item_pairs = remove_duplicate_user_item_pairs
        self.num_users = num_users
        self.num_items = num_items

        self.num_interactions = self.mat.nnz
        self.min_rating = self.mat.data.min()
        self.max_rating = self.mat.data.max()

    @abstractmethod
    def __getitem__(self, index: Union[int, Iterable[int]]) -> (
        Union[Tuple[Tuple[int, int], np.array], Tuple[Tuple[np.array, np.array], np.array]]
    ):
        """Access item in the ``BaseInteractions`` instance."""
        pass

    def __len__(self) -> int:
        """Number of non-zero interactions in the ``BaseInteractions`` instance."""
        return self.num_interactions

    def todense(self) -> np.matrix:
        """Transforms ``BaseInteractions`` instance sparse matrix to np.matrix, 2-d."""
        return self.mat.todense()

    def toarray(self) -> np.array:
        """Transforms ``BaseInteractions`` instance sparse matrix to np.array, 2-d."""
        return self.mat.toarray()

    def head(self, n: int = 5) -> np.array:
        """Return the first ``n`` rows of the dense matrix as a np.array, 2-d."""
        n = self._prep_head_tail_n(n=n)
        return self.mat.tocsr()[range(n), :].toarray()

    def tail(self, n: int = 5) -> np.array:
        """Return the last ``n`` rows of the dense matrix as a np.array, 2-d."""
        n = self._prep_head_tail_n(n=n)
        return self.mat.tocsr()[range(-n, 0), :].toarray()

    def _prep_head_tail_n(self, n: int) -> int:
        """Ensure we don't run into an ``IndexError`` when using ``head`` or ``tail`` methods."""
        if n < 0:
            n = self.num_users + n
        if n > self.num_users:
            n = self.num_users

        return n


class Interactions(BaseInteractions):
    """
    PyTorch ``Dataset`` for implicit user-item interactions data.

    If ``mat`` is provided, the ``Interactions`` instance will act as a wrapper for a sparse matrix
    in COOrdinate format, typically looking like:

    * Users comprising the rows

    * Items comprising the columns

    * Ratings given by that user for that item comprising the elements of the matrix

    ``Interactions`` can be instantiated instead by passing in single arrays with corresponding
    user_ids, item_ids, and ratings (by default, set to 1 for implicit recommenders) values with
    the same functionality as a matrix. Note that with this approach, the number of users and items
    will be the maximum values in those two columns, respectively, and it is expected that all
    integers between 0 and the maximum ID should appear somewhere in the data.

    By default, exact negative sampling will be used during each ``__getitem__`` call. To use
    approximate negative sampling, set ``max_number_of_samples_to_consider = 0``. This will avoid
    building a positive item lookup dictionary during initialization.

    Unlike in ``ExplicitInteractions``, we rely on negative sampling for implicit data. Each
    ``__getitem__`` call will thus return a nested tuple containing user IDs, item IDs, and
    sampled negative item IDs. This nested vs. non-nested structure is key for the model to
    determine where it should be implicit or explicit. Use the table below for reference:

    .. list-table::
        :header-rows: 1

        * - ``__getitem__`` Format
          - Expected Meaning
          - Model Type
        * - ``((X, Y), Z)``
          - ``((user IDs, item IDs), negative item IDs)``
          - **Implicit**
        * - ``(X, Y, Z)``
          - ``(user IDs, item IDs, ratings)``
          - **Explicit**

    Parameters
    -------------
    mat: scipy.sparse.coo_matrix or numpy.array, 2-dimensional
        Interactions matrix, which, if provided, will be used instead of ``users``, ``items``, and
        ``ratings`` arguments
    users: Iterable[int], 1-d
        Array of user IDs, starting at 0
    items: Iterable[int], 1-d
        Array of corresponding item IDs to ``users``, starting at 0
    ratings: Iterable[int], 1-d
        Array of corresponding ratings to both ``users`` and ``items``. If ``None``, will default to
        each user in ``user`` interacting with an item with a rating value of 1
    num_negative_samples: int
        Number of negative samples to return with each ``__getitem__`` call
    allow_missing_ids: bool
        If ``False``, will check that both ``users`` and ``items`` contain each integer from 0 to
        the maximum value in the array. This check only applies when initializing an
        ``Interactions`` instance using 1-dimensional arrays ``users`` and ``items``
    remove_duplicate_user_item_pairs: bool
        Will check for and remove any duplicate user, item ID pairs from the ``Interactions`` matrix
        during initialization. Note that this will create a second sparse matrix held in memory
        to efficiently check, which could cause memory concerns for larger data. If you are sure
        that there are no duplicated, user, item ID pairs, set to ``False``
    num_users: int
        Number of users in the dataset. If ``num_users == 'infer'``, this will be set to the
        ``mat.shape[0]`` or ``max(users) + 1``, depending on the input
    num_items: int
        Number of items in the dataset. If ``num_items == 'infer'``, this will be set to the
        ``mat.shape[1]`` or ``max(items) + 1``, depending on the input
    check_num_negative_samples_is_valid: bool
        Check that ``num_negative_samples`` is less than the maximum number of items a user has
        interacted with. If it is not, then for all users who have fewer than
        ``num_negative_samples`` items not interacted with, a random sample including positive items
        will be returned as negative
    max_number_of_samples_to_consider: int
        Number of samples to try for a given user before returning an approximate negative sample.
        This should be greater than ``num_negative_samples``. If set to ``0``, approximate negative
        sampling will be used by default in ``__getitem__`` and a positive item lookup dictionary
        will NOT be built
    seed: int
        Seed for random sampling

    """
    def __init__(self,
                 mat: Optional[Union[coo_matrix, np.array]] = None,
                 users: Optional[Iterable[int]] = None,
                 items: Optional[Iterable[int]] = None,
                 ratings: Optional[Iterable[int]] = None,
                 num_negative_samples: int = 10,
                 allow_missing_ids: bool = False,
                 remove_duplicate_user_item_pairs: bool = True,
                 num_users: int = 'infer',
                 num_items: int = 'infer',
                 check_num_negative_samples_is_valid: bool = True,
                 max_number_of_samples_to_consider: int = 200,
                 seed: Optional[int] = None):
        if mat is None and ratings is not None and 0 in set(ratings):
            warnings.warn(
                '``ratings`` contain ``0``s, which are ignored for implicit data.'
                ' Filtering these rows out.'
            )
            indices_to_drop = [idx for idx, rating in enumerate(ratings) if rating == 0]

            users = _drop_array_values_by_idx(array=users, indices_to_drop=indices_to_drop)
            items = _drop_array_values_by_idx(array=items, indices_to_drop=indices_to_drop)
            ratings = _drop_array_values_by_idx(array=ratings, indices_to_drop=indices_to_drop)

        super().__init__(mat=mat,
                         users=users,
                         items=items,
                         ratings=ratings,
                         allow_missing_ids=allow_missing_ids,
                         remove_duplicate_user_item_pairs=remove_duplicate_user_item_pairs,
                         num_users=num_users,
                         num_items=num_items)

        if seed is None:
            seed = collie.utils.get_random_seed()

        self.num_negative_samples = num_negative_samples
        self.max_number_of_samples_to_consider = max_number_of_samples_to_consider
        self.check_num_negative_samples_is_valid = check_num_negative_samples_is_valid
        self.seed = seed

        random.seed(self.seed)

        assert self.num_negative_samples >= 1

        if (
            self.num_negative_samples >= self.max_number_of_samples_to_consider
            and self.max_number_of_samples_to_consider > 0
        ):
            # no warning for ``max_number_of_samples_to_consider==0`` since it is likely intentional
            warnings.warn(
                '``num_negative_samples > max_number_of_samples_to_consider``. Approximate negative'
                ' sampling will be used.'
            )

        # When an ``Interactions`` is instantiated with exact negative sampling, a
        # ``positive_items`` attribute is created, a ``set`` of the ``mat`` object that enables
        # fast, O(1), ``(row, col)`` lookup. When ``__getitem__`` is called, negative item IDs are
        # sampled one-at-a-time from all possible values in ``self.num_items``, we check if that
        # user ID, item ID pair is in ``self.positive_items``, and sample continuously until we
        # have a negative match or reach a limit of ``max_number_of_samples_to_consider`` tries
        if self.check_num_negative_samples_is_valid:
            print('Checking ``num_negative_samples`` is valid...')
            counter = collections.Counter(self.mat.row)
            max_number_of_items_interacted_with = counter.most_common(1)[0][1]
            print('Maximum number of items a user has interacted with: {}'.format(
                max_number_of_items_interacted_with
            ))

            del counter

            is_valid = (
                self.num_negative_samples
                < (self.num_items - max_number_of_items_interacted_with)
            )
            assert is_valid, '``num_negative_samples`` must be less than {}!'.format(
                (self.num_items - max_number_of_items_interacted_with)
            )

        self.positive_items = {}
        if self.max_number_of_samples_to_consider > 0:
            print('Generating positive items set...')
            self._generate_positive_item_set()

    def _generate_positive_item_set(self) -> None:
        """Build positive item dictionary lookup for exact negative sampling."""
        self.positive_items = set(zip(self.mat.row, self.mat.col))

    def __repr__(self) -> str:
        """String representation of ``Interactions`` class."""
        return textwrap.dedent(
            f'''
            Interactions object with {self.num_interactions} interactions between {self.num_users}
            users and {self.num_items} items, returning {self.num_negative_samples} negative
            samples per interaction.
            '''
        ).replace('\n', ' ').strip()

    def __getitem__(self, index: Union[int, Iterable[int]]) -> (
        Union[Tuple[Tuple[int, int], np.array], Tuple[Tuple[np.array, np.array], np.array]]
    ):
        """Access item in the ``Interactions`` instance, returning negative samples as well."""
        user_id = self.mat.row[index]
        item_id = self.mat.col[index]
        # rating = self.mat.data[index]  # not needed for any loss currently implemented

        negative_item_ids_array = self._negative_sample(user_id)

        return (user_id, item_id), negative_item_ids_array

    def _negative_sample(self, user_id: Union[int, np.array]) -> np.array:
        """Generate negative samples for a ``user_id``."""
        if self.max_number_of_samples_to_consider > 0:
            # if we are here, we are doing true negative sampling
            negative_item_ids_list = list()

            if not isinstance(user_id, collections.abc.Iterable):
                user_id = [user_id]

            for specific_user_id in user_id:
                # generate true negative samples for the ``user_id``
                samples_checked = 0
                temp_negative_item_ids_list = list()

                while len(temp_negative_item_ids_list) < self.num_negative_samples:
                    negative_item_id = random.choice(range(self.num_items))
                    # we have a negative sample, make sure the user has not interacted with it
                    # before, else we resample and try again
                    while (
                        (specific_user_id, negative_item_id) in self.positive_items
                        or negative_item_id in temp_negative_item_ids_list
                    ):
                        if samples_checked >= self.max_number_of_samples_to_consider:
                            num_samples_left_to_generate = (
                                self.num_negative_samples - len(temp_negative_item_ids_list) - 1
                            )
                            temp_negative_item_ids_list += random.choices(
                                range(self.num_items), k=num_samples_left_to_generate
                            )
                            break

                        negative_item_id = random.choice(range(self.num_items))
                        samples_checked += 1

                    temp_negative_item_ids_list.append(negative_item_id)

                negative_item_ids_list += [np.array(temp_negative_item_ids_list)]

            if len(user_id) > 1:
                negative_item_ids_array = np.stack(negative_item_ids_list)
            else:
                negative_item_ids_array = negative_item_ids_list[0]
        else:
            # if we are here, we are doing approximate negative sampling
            if isinstance(user_id, collections.abc.Iterable):
                size = (len(user_id), self.num_negative_samples)
            else:
                size = (self.num_negative_samples,)

            negative_item_ids_array = np.random.randint(
                low=0,
                high=self.num_items,
                size=size,
            )

        return negative_item_ids_array


class ExplicitInteractions(BaseInteractions):
    """
    PyTorch ``Dataset`` for explicit user-item interactions data.

    If ``mat`` is provided, the ``Interactions`` instance will act as a wrapper for a sparse matrix
    in COOrdinate format, typically looking like:

    * Users comprising the rows

    * Items comprising the columns

    * Ratings given by that user for that item comprising the elements of the matrix

    ``Interactions`` can be instantiated instead by passing in single arrays with corresponding
    user_ids, item_ids, and ratings values with the same functionality as a matrix. Note that with
    this approach, the number of users and items will be the maximum values in those two columns,
    respectively, and it is expected that all integers between 0 and the maximum ID should appear
    somewhere in the user or item ID data.

    Unlike in ``Interactions``, there is no need for negative sampling for explicit data. Each
    ``__getitem__`` call will thus return a single, non-nested tuple containing user IDs, item IDs,
    and ratings. This nested vs. non-nested structure is key for the model to determine where it
    should be implicit or explicit. Use the table below for reference:

    .. list-table::
        :header-rows: 1

        * - ``__getitem__`` Format
          - Expected Meaning
          - Model Type
        * - ``((X, Y), Z)``
          - ``((user IDs, item IDs), negative item IDs)``
          - **Implicit**
        * - ``(X, Y, Z)``
          - ``(user IDs, item IDs, ratings)``
          - **Explicit**

    Parameters
    -------------
    mat: scipy.sparse.coo_matrix or numpy.array, 2-dimensional
        Interactions matrix, which, if provided, will be used instead of ``users``, ``items``, and
        ``ratings`` arguments
    users: Iterable[int], 1-d
        Array of user IDs, starting at 0
    items: Iterable[int], 1-d
        Array of corresponding item IDs to ``users``, starting at 0
    ratings: Iterable[int], 1-d
        Array of corresponding ratings to both ``users`` and ``items``. If ``None``, will default to
        each user in ``user`` interacting with an item with a rating value of 1
    allow_missing_ids: bool
        If ``False``, will check that both ``users`` and ``items`` contain each integer from 0 to
        the maximum value in the array. This check only applies when initializing an
        ``ExplicitInteractions`` instance using 1-dimensional arrays ``users`` and ``items``
    remove_duplicate_user_item_pairs: bool
        Will check for and remove any duplicate user, item ID pairs from the
        ``ExplicitInteractions`` matrix during initialization. Note that this will create a second
        sparse matrix held in memory to efficiently check, which could cause memory concerns for
        larger data. If you are sure that there are no duplicated, user, item ID pairs, set to
        ``False``
    num_users: int
        Number of users in the dataset. If ``num_users == 'infer'``, this will be set to the
        ``mat.shape[0]`` or ``max(users) + 1``, depending on the input
    num_items: int
        Number of items in the dataset. If ``num_items == 'infer'``, this will be set to the
        ``mat.shape[1]`` or ``max(items) + 1``, depending on the input

    """
    def __init__(self,
                 mat: Optional[Union[coo_matrix, np.array]] = None,
                 users: Optional[Iterable[int]] = None,
                 items: Optional[Iterable[int]] = None,
                 ratings: Optional[Iterable[int]] = None,
                 allow_missing_ids: bool = False,
                 remove_duplicate_user_item_pairs: bool = True,
                 num_users: int = 'infer',
                 num_items: int = 'infer'):
        if mat is None and ratings is None:
            raise ValueError(
                'Ratings must be provided to ``ExplicitInteractions`` with ``mat`` or ``ratings``'
                ' - both cannot be ``None``!'
            )

        super().__init__(mat=mat,
                         users=users,
                         items=items,
                         ratings=ratings,
                         allow_missing_ids=allow_missing_ids,
                         remove_duplicate_user_item_pairs=remove_duplicate_user_item_pairs,
                         num_users=num_users,
                         num_items=num_items)

    @property
    def num_negative_samples(self) -> int:
        """Does not exist for explicit data."""
        raise AttributeError('``num_negative_samples`` does not exist for explicit datasets.')

    def __repr__(self) -> str:
        """String representation of ``ExplicitInteractions`` class."""
        return textwrap.dedent(
            f'''
            ExplicitInteractions object with {self.num_interactions} interactions between
            {self.num_users} users and {self.num_items} items, with minimum rating of
            {self.min_rating} and maximum rating of {self.max_rating}.
            '''
        ).replace('\n', ' ').strip()

    def __getitem__(self, index: Union[int, Iterable[int]]) -> (
        Union[Tuple[int, int, np.array], Tuple[np.array, np.array, np.array]]
    ):
        """Access item in the ``ExplicitInteractions`` instance."""
        user_id = self.mat.row[index]
        item_id = self.mat.col[index]
        rating = self.mat.data[index]

        return user_id, item_id, rating


class HDF5Interactions(torch.utils.data.Dataset):
    """
    Create an ``Interactions``-like object for data in the HDF5 format that might be too large to
    fit in memory.

    Many of the same features of ``Interactions`` are implemented here, with the exception that
    approximate negative sampling will always be used.

    Parameters
    ----------
    hdf5_path: str
    user_col: str
        Column in HDF5 file with user IDs. IDs must begin at 0
    item_col: str
        Column in HDF5 file with item IDs. IDs must begin at 0
    num_negative_samples: int
        Number of negative samples to return with each ``__getitem__`` call
    num_users: int
        Number of users in the dataset. If ``num_users == 'infer'`` and there is not a ``meta`` key
        in ``hdf5_path``'s HDF5 dataset, this will be set to the the maximum value in
        ``user_col`` + 1, found by iterating through the entire dataset
    num_items: int
        Number of items in the dataset. If ``num_items == 'infer'`` and there is not an ``meta`` key
        in ``hdf5_path``'s HDF5 dataset, this will be set to the the maximum value in
        ``item_col`` + 1, found by iterating through the entire dataset
    seed: int
        Seed for random sampling and shuffling if ``shuffle is True``
    shuffle: bool
        Shuffle data in a batch. For example, if one calls ``__getitem__`` with
        ``start_idx_and_batch_size = (0, 4)`` and ``shuffle is False``, this will always return the
        data at indices 0, 1, 2, 3 in order. However, the same call with ``shuffle = True`` will
        return a random shuffle of 0, 1, 2, 3 each call. This is recommended for use in a
        ``HDF5InteractionsDataLoader`` for training data in lieu of true data shuffling

    """
    def __init__(self,
                 hdf5_path: str,
                 user_col: str = 'users',
                 item_col: str = 'items',
                 num_negative_samples: int = 10,
                 num_users: int = 'infer',
                 num_items: int = 'infer',
                 seed: Optional[int] = None,
                 shuffle: bool = False):
        self.hdf5_path = hdf5_path
        self.user_col = user_col
        self.item_col = item_col
        self.num_negative_samples = num_negative_samples
        self.seed = seed
        self.shuffle = shuffle

        with pd.HDFStore(self.hdf5_path, mode='r', complib='blosc') as store:
            self.num_interactions = store.get_storer('interactions').shape

            if isinstance(num_users, int) and isinstance(num_items, int):
                self.num_users = num_users
                self.num_items = num_items
            else:
                try:
                    chunk = store.select('meta')
                    self.num_users = chunk['num_users'].item()
                    self.num_items = chunk['num_items'].item()
                except KeyError:
                    print('``meta`` key not found - generating ``num_users`` and ``num_items``.')

                    self.num_users = 0
                    self.num_items = 0
                    # while we are here, we can also check minimum IDs are 0 for free
                    # TODO: is there a more efficient way to check this? should we always check?
                    min_user_id = 1
                    min_item_id = 1

                    # default Pandas ``chunksize`` is 100000, so we will use that too
                    chunksize = 100000
                    for idx in tqdm(range(0, self.num_interactions, chunksize)):
                        chunk = store.select('interactions', start=idx, stop=(idx + chunksize))
                        self.num_users = max(chunk[self.user_col].max(), self.num_users)
                        self.num_items = max(chunk[self.item_col].max(), self.num_items)
                        min_user_id = min(chunk[self.user_col].min(), min_user_id)
                        min_item_id = min(chunk[self.item_col].min(), min_item_id)

                    if min_user_id != 0 or min_item_id != 0:
                        raise ValueError(
                            f'Minimum values of {user_col} and {item_col} in HDF5 data must both be'
                            f' 0, not {min_user_id} and {min_item_id}, respectively.'
                        )

                    # add one here since ``users`` and ``items`` are both zero-indexed
                    self.num_users += 1
                    self.num_items += 1

        assert self.num_users > 1
        assert self.num_items > 1

        if self.seed is None:
            self.seed = collie.utils.get_random_seed()

        np.random.seed(seed=self.seed)

    def __getitem__(self, start_idx_and_batch_size: Tuple[int, int]) -> (
        Tuple[Tuple[np.array, np.array], np.array]
    ):
        """Get a batch of data."""
        if isinstance(start_idx_and_batch_size, tuple):
            start_idx, batch_size = start_idx_and_batch_size
        else:
            start_idx = start_idx_and_batch_size
            batch_size = 1

        chunk = self._get_data_chunk(start_idx, batch_size)

        if len(chunk) == 0:
            raise IndexError(f'Index {start_idx} out of range for HDF5 data.')

        user_ids = chunk[self.user_col].to_numpy()
        item_ids = chunk[self.item_col].to_numpy()
        # ratings = chunk[self.ratings_col].to_numpy()  # not needed for any implemented loss yet

        if self.shuffle:
            idxs = np.random.permutation(len(user_ids))
            user_ids = user_ids[idxs]
            item_ids = item_ids[idxs]

        negative_item_ids = np.random.randint(
            low=0,
            high=self.num_items,
            size=(len(user_ids), self.num_negative_samples)
        )

        return (user_ids, item_ids), negative_item_ids

    def _get_data_chunk(self, start_idx: int, batch_size: int) -> pd.DataFrame:
        with pd.HDFStore(self.hdf5_path, mode='r', complib='blosc') as store:
            return store.select('interactions',
                                start=start_idx,
                                stop=(start_idx + batch_size))

    def __len__(self) -> int:
        """Get number of batches."""
        return self.num_interactions

    def __repr__(self) -> str:
        """String representation of ``HDF5Interactions`` class."""
        return textwrap.dedent(
            f'''
            HDF5Interactions object with {self.num_interactions} interactions between
            {self.num_users} users and {self.num_items} items, returning
            {self.num_negative_samples} negative samples per interaction.
            '''
        ).replace('\n', ' ').strip()

    def head(self, n: int = 5) -> pd.DataFrame:
        """Return the first ``n`` rows of the underlying pd.DataFrame."""
        n = self._prep_head_tail_n(n=n)
        return self._get_data_chunk(0, n)

    def tail(self, n: int = 5) -> pd.DataFrame:
        """Return the last ``n`` rows of the underlying pd.DataFrame."""
        n = self._prep_head_tail_n(n=n)
        return self._get_data_chunk(self.num_interactions - n, n)

    def _prep_head_tail_n(self, n: int) -> int:
        """Ensure we don't run into an ``IndexError`` when using ``head`` or ``tail`` methods."""
        if n < 0:
            n = self.num_interactions + n
        if n > self.num_interactions:
            n = self.num_interactions

        return n


def _check_array_contains_all_integers(array: Iterable[int],
                                       array_max_value: int,
                                       array_name: str = 'Array') -> None:
    """Check that an array has all numbers between 0 and ``array_max``."""
    if set(array) != set(range(array_max_value)):
        raise ValueError(
            f'``{array_name}`` must contain every integer between 0 and {array_max_value - 1}. '
            + 'To override this error, set ``allow_missing_ids`` to True.'
        )


def _drop_array_values_by_idx(array: Iterable[Any], indices_to_drop: Iterable[int]) -> List[Any]:
    return [element for idx, element in enumerate(array) if idx not in indices_to_drop]



from collie.model.base import *
from collie.model.cold_start_matrix_factorization import *
from collie.model.collaborative_metric_learning import *
from collie.model.deep_fm import *
from collie.model.hybrid_matrix_factorization import *
from collie.model.hybrid_pretrained_matrix_factorization import *
from collie.model.matrix_factorization import *
from collie.model.mlp_matrix_factorization import *
from collie.model.neural_collaborative_filtering import *
from collie.model.nonlinear_matrix_factorization import *



import multiprocessing
import textwrap
from typing import Iterable, Optional, Union

import numpy as np
from scipy.sparse import coo_matrix
import torch

from collie.interactions.datasets import (BaseInteractions,
                                          ExplicitInteractions,
                                          HDF5Interactions,
                                          Interactions)
from collie.interactions.samplers import ApproximateNegativeSampler, HDF5Sampler


class BaseInteractionsDataLoader(torch.utils.data.DataLoader):
    """
    A base class acting as a wrapper around a ``torch.utils.data.DataLoader`` for
    ``Interactions``-type datasets. This class should only be inherited from and not used for model
    training.

    Parameters
    ----------
    interactions: Interactions or HDF5Interactions
    num_workers: int
        Number of subprocesses to use for data loading
    **kwargs: keyword arguments
        Keyword arguments passed into ``torch.utils.data.DataLoader.__init__``:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader

    """
    def __init__(self,
                 interactions: Union[Interactions, HDF5Interactions] = None,
                 num_workers: int = multiprocessing.cpu_count(),
                 **kwargs):

        super().__init__(
            dataset=interactions,
            num_workers=num_workers,
            **kwargs,
        )

        self.interactions = interactions

    @property
    def num_users(self) -> int:
        """Number of users in ``interactions``."""
        return self.interactions.num_users

    @property
    def num_items(self) -> int:
        """Number of items in ``interactions``."""
        return self.interactions.num_items

    @property
    def num_negative_samples(self) -> int:
        """Number of negative samples in ``interactions``."""
        return self.interactions.num_negative_samples

    @property
    def num_interactions(self) -> int:
        """Number of interactions in ``interactions``."""
        return self.interactions.num_interactions

    @property
    def mat(self) -> coo_matrix:
        """Sparse COO matrix of ``interactions``."""
        return self.interactions.mat


class InteractionsDataLoader(BaseInteractionsDataLoader):
    """
    A light wrapper around a ``torch.utils.data.DataLoader`` for ``Interactions``-type datasets.

    For implicit data, batches will be created one-point-at-a-time using exact negative sampling
    (unless configured not to in ``interactions``), which is optimal when datasets are smaller
    (< 1M+ interactions) and model training speed is not a concern. This is the default
    ``DataLoader`` for ``Interactions`` datasets.

    For explicit data, negative sampling is not used, but batches will still be created
    one-point-at-a-time.

    Parameters
    ----------
    interactions: BaseInteractions
        If not provided, an ``Interactions`` object will be created with ``mat`` or all of
        ``users``, ``items``, and ``ratings``
    mat: scipy.sparse.coo_matrix or numpy.array, 2-dimensional
        If ``interactions is None``, will be used instead of ``users``, ``items``, and ``ratings``
        arguments to create an ``Interactions`` object
    users: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of user IDs, starting at 0
    items: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of corresponding item IDs to ``users``,
        starting at 0
    ratings: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of corresponding ratings to both
        ``users`` and ``items``. If ``None``, will default to each user in ``user`` interacting with
        an item with a rating value of 1
    batch_size: int
        Number of samples per batch to load
    shuffle: bool
        Whether to shuffle the order of data returned or not. This is especially useful for training
        data to ensure the model does not overfit to a specific order of data
    num_workers: int
        Number of subprocesses to use for data loading
    **kwargs: keyword arguments
        Relevant keyword arguments will be passed into ``Interactions`` object creation, if
        ``interactions is None`` and the keyword argument matches one of
        ``Interactions.__init__.__code__.co_varnames``. All other keyword arguments will be passed
        into ``torch.utils.data.DataLoader``:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader

    Attributes
    ----------
    interactions: Interactions (default) or ExplicitInteractions

    """
    def __init__(self,
                 interactions: BaseInteractions = None,
                 mat: Optional[Union[coo_matrix, np.array]] = None,
                 users: Optional[Iterable[int]] = None,
                 items: Optional[Iterable[int]] = None,
                 ratings: Optional[Iterable[int]] = None,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 num_workers: int = multiprocessing.cpu_count(),
                 **kwargs):
        if interactions is None:
            # find all kwargs in the ``__init__`` for a ``Interactions`` object
            interactions_only_kwargs = {
                k: v for k, v in kwargs.items() if k in Interactions.__init__.__code__.co_varnames
            }
            # find all kwargs not in the ``__init__`` for a ``Interactions`` object OR all kwargs
            # that are used in the ``__init__`` for a ``torch.utils.data.DataLoader`` object
            kwargs = {
                k: v for k, v in kwargs.items()
                if k not in Interactions.__init__.__code__.co_varnames
                or k in torch.utils.data.DataLoader.__init__.__code__.co_varnames
            }

            interactions = Interactions(mat=mat,
                                        users=users,
                                        items=items,
                                        ratings=ratings,
                                        **interactions_only_kwargs)

        super().__init__(
            interactions=interactions,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            **kwargs,
        )

        self.shuffle = shuffle

    def __repr__(self) -> str:
        """String representation of ``InteractionsDataLoader`` class."""
        if hasattr(self.interactions, 'num_negative_samples'):
            extra_repr_str = (
                f'{self.num_negative_samples} negative samples per implicit interaction in'
            )
        else:
            extra_repr_str = 'explicit,'

        return textwrap.dedent(
            f'''
            InteractionsDataLoader object with {self.num_interactions} interactions between
            {self.num_users} users and {self.num_items} items, returning
            {extra_repr_str}
            {'shuffled' if self.shuffle else 'non-shuffled'} batches of size {self.batch_size}.
            '''
        ).replace('\n', ' ').strip()


class ApproximateNegativeSamplingInteractionsDataLoader(BaseInteractionsDataLoader):
    """
    A computationally more efficient ``DataLoader`` for ``Interactions`` data using approximate
    negative sampling for negative items.

    This DataLoader groups ``__getitem__`` calls together into a single operation, which
    dramatically speeds up a traditional DataLoader's process of calling ``__getitem__`` one index
    at a time, then concatenating them together before returning. In an effort to batch operations
    together, all negative samples returned will be approximate, meaning this does not check if a
    user has previously interacted with the item. With a sufficient number of interactions (1M+), we
    have found a speed increase of 2x at the cost of a 1% reduction in MAP @ 10 performance
    compared to ``InteractionsDataLoader``.

    For greater efficiency, we disable automated batching by setting the DataLoader's
    ``batch_size`` attribute to ``None``. Thus,
    to access the "true" batch size that the sampler uses, access
    ``ApproximateNegativeSamplingInteractionsDataLoader.approximate_negative_sampler.batch_size``.

    Parameters
    ----------
    interactions: Interactions
        If not provided, an ``Interactions`` object will be created with ``mat`` or all of
        ``users``, ``items``, and ``ratings`` with ``max_number_of_samples_to_consider=0``
    mat: scipy.sparse.coo_matrix or numpy.array, 2-dimensional
        If ``interactions is None``, will be used instead of ``users``, ``items``, and ``ratings``
        arguments to create an ``Interactions`` object
    users: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of user IDs, starting at 0
    items: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of corresponding item IDs to ``users``,
        starting at 0
    ratings: Iterable[int], 1-d
        If ``interactions is None and mat is None``, array of corresponding ratings to both
        ``users`` and ``items``. If ``None``, will default to each user in ``user`` interacting with
        an item with a rating value of 1
    batch_size: int
        Number of samples per batch to load
    shuffle: bool
        Whether to shuffle the order of data returned or not. This is especially useful for training
        data to ensure the model does not overfit to a specific order of data
    **kwargs: keyword arguments
        Relevant keyword arguments will be passed into ``Interactions`` object creation, if
        ``interactions is None`` and the keyword argument matches one of
        ``Interactions.__init__.__code__.co_varnames``. All other keyword arguments will be passed
        into ``torch.utils.data.DataLoader``:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader

    Attributes
    ----------
    interactions: Interactions

    """
    def __init__(self,
                 interactions: Interactions = None,
                 mat: Optional[Union[coo_matrix, np.array]] = None,
                 users: Optional[Iterable[int]] = None,
                 items: Optional[Iterable[int]] = None,
                 ratings: Optional[Iterable[int]] = None,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 num_workers: int = multiprocessing.cpu_count(),
                 **kwargs):
        if isinstance(interactions, ExplicitInteractions):
            raise ValueError(
                '``ApproximateNegativeSamplingInteractionsDataLoader`` does not support explicit'
                ' data types!'
            )

        if interactions is None:
            interactions_only_kwargs = {
                k: v for k, v in kwargs.items()
                if k in Interactions.__init__.__code__.co_varnames
            }
            kwargs = {
                k: v for k, v in kwargs.items()
                if k not in Interactions.__init__.__code__.co_varnames
                or k in torch.utils.data.DataLoader.__init__.__code__.co_varnames
            }

            interactions = Interactions(mat=mat,
                                        users=users,
                                        items=items,
                                        ratings=ratings,
                                        max_number_of_samples_to_consider=0,
                                        **interactions_only_kwargs)
        else:
            # we need ``max_number_of_samples_to_consider`` to be 0 in order to do approximate
            # negative sampling
            interactions.max_number_of_samples_to_consider = 0

        approximate_negative_sampler = ApproximateNegativeSampler(interactions=interactions,
                                                                  batch_size=batch_size,
                                                                  shuffle=shuffle,
                                                                  seed=interactions.seed)

        super().__init__(
            interactions=interactions,
            sampler=approximate_negative_sampler,
            num_workers=num_workers,
            batch_size=None,  # Disable automated batching
            **kwargs,
        )

        self.approximate_negative_sampler = approximate_negative_sampler
        self.shuffle = shuffle

    def __repr__(self) -> str:
        """String representation of ``ApproximateNegativeSamplingInteractionsDataLoader`` class."""
        return textwrap.dedent(
            f'''
            ApproximateNegativeSamplingInteractionsDataLoader object with {self.num_interactions}
            interactions between {self.num_users} users and {self.num_items} items, returning
            {self.num_negative_samples} negative samples per implicit interaction in
            {'shuffled' if self.shuffle else 'non-shuffled'} batches of size
            {self.approximate_negative_sampler.batch_size}.
            '''
        ).replace('\n', ' ').strip()


class HDF5InteractionsDataLoader(BaseInteractionsDataLoader):
    """
    A light wrapper around a ``torch.utils.data.DataLoader`` for HDF5 data, with behavior very
    similar to ``ApproximateNegativeSamplingInteractionsDataLoader``.

    If not provided, a ``HDF5Interactions`` dataset will be created as the data for the
    ``DataLoader``. A custom sampler, ``HDF5Sampler``, will also be instantiated for the
    ``DataLoader`` to use that allows sampling in batches that make for faster HDF5 data reads from
    disk.

    While similar to a standard ``DataLoader``, note that when ``shuffle is True``, this will only
    shuffle the order of batches and the data within batches to still make for efficient reading
    of HDF5 data from disk, rather than shuffling across the entire dataset.

    For greater efficiency, we disable automated batching by setting the DataLoader's
    ``batch_size`` attribute to ``None``. Thus,
    to access the "true" batch size that the sampler uses, access
    ``HDF5InteractionsDataLoader.hdf5_sampler.batch_size``.

    Parameters
    ----------
    hdf5_interactions: HDF5Interactions
        If provided, will override input argument for ``hdf5_path``
    hdf5_path: str
        If ``hdf5_interactions is None``, the path to the HDF5 dataset
    batch_size: int
        Number of samples per batch to load
    shuffle: bool
        Whether to shuffle the order of batches returned or not. This is especially useful for
        training data to ensure the model does not overfit to a specific order of data. Note that
        this will not perform a true shuffle of the data, but shuffle the order of batches. While
        this is an approximation of true sampling, it allows us a greater speed up during model
        training for a negligible effect on model performance
    num_workers: int
        Number of subprocesses to use for data loading
    **kwargs: keyword arguments
        Relevant keyword arguments will be passed into ``HDF5Interactions`` object creation, if
        ``hdf5_interactions is None`` and the keyword argument matches one of
        ``HDF5Interactions.__init__.__code__.co_varnames``. All other keyword arguments will be
        passed into ``torch.utils.data.DataLoader``:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader

    """
    def __init__(self,
                 hdf5_interactions: HDF5Interactions = None,
                 hdf5_path: Optional[str] = None,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 num_workers: int = multiprocessing.cpu_count(),
                 **kwargs):
        if hdf5_interactions is None:
            # find all kwargs in the ``__init__`` for a ``HDF5Interactions`` object
            interactions_only_kwargs = {
                k: v for k, v in kwargs.items()
                if k in HDF5Interactions.__init__.__code__.co_varnames
            }
            # find all kwargs not in the ``__init__`` for a ``HDF5Interactions`` object OR all
            # kwargs that are used in the ``__init__`` for a ``torch.utils.data.DataLoader`` object
            kwargs = {
                k: v for k, v in kwargs.items()
                if k not in HDF5Interactions.__init__.__code__.co_varnames
                or k in torch.utils.data.DataLoader.__init__.__code__.co_varnames
            }

            hdf5_interactions = HDF5Interactions(hdf5_path=hdf5_path,
                                                 **interactions_only_kwargs)

        hdf5_sampler = HDF5Sampler(hdf5_interactions=hdf5_interactions,
                                   batch_size=batch_size,
                                   shuffle=shuffle,
                                   seed=hdf5_interactions.seed)

        super().__init__(
            interactions=hdf5_interactions,
            sampler=hdf5_sampler,
            num_workers=num_workers,
            batch_size=None,  # Disable automated batching
            **kwargs,
        )

        self.hdf5_sampler = hdf5_sampler
        self.hdf5_path = hdf5_path
        self.shuffle = shuffle

    @property
    def mat(self) -> None:
        """``mat`` attribute is not possible to access in ``HDF5InteractionsDataLoader``."""
        raise AttributeError('``HDF5InteractionsDataLoader`` cannot support ``mat`` attribute since'
                             ' data is read in from disk dynamically.')

    def __repr__(self) -> str:
        """String representation of ``HDF5InteractionsDataLoader`` class."""
        return textwrap.dedent(
            f'''
            HDF5InteractionsDataLoader object with {self.interactions.num_interactions}
            interactions between {self.interactions.num_users} users and
            {self.interactions.num_items} items, returning {self.num_negative_samples} negative
            samples per implicit interaction in {'shuffled' if self.shuffle else 'non-shuffled'}
            batches of size {self.hdf5_sampler.batch_size}.
            '''
        ).replace('\n', ' ').strip()



import pytest
from sklearn.model_selection import train_test_split

from collie.cross_validation import random_split, stratified_split
from collie.interactions import ExplicitInteractions, Interactions
from collie.movielens import (get_movielens_metadata,
                              read_movielens_df,
                              read_movielens_df_item,
                              read_movielens_posters_df)
from collie.utils import convert_to_implicit


@pytest.fixture(scope='session')
def movielens_explicit_df():
    return read_movielens_df(decrement_ids=True)


@pytest.fixture(scope='session')
def movielens_explicit_df_not_decremented():
    return read_movielens_df(decrement_ids=False)


@pytest.fixture(scope='session')
def movielens_implicit_df(movielens_explicit_df):
    return convert_to_implicit(movielens_explicit_df)


@pytest.fixture(scope='session')
def movielens_df_item():
    return read_movielens_df_item()


@pytest.fixture(scope='session')
def movielens_posters_df():
    return read_movielens_posters_df()


@pytest.fixture(scope='session')
def movielens_metadata_df():
    return get_movielens_metadata()


@pytest.fixture(scope='session')
def movielens_implicit_interactions(movielens_implicit_df):
    return Interactions(users=movielens_implicit_df['user_id'],
                        items=movielens_implicit_df['item_id'],
                        ratings=movielens_implicit_df['rating'],
                        num_negative_samples=10,
                        max_number_of_samples_to_consider=200,
                        allow_missing_ids=True)


@pytest.fixture(scope='session')
def movielens_explicit_interactions(movielens_explicit_df):
    return ExplicitInteractions(users=movielens_explicit_df['user_id'],
                                items=movielens_explicit_df['item_id'],
                                ratings=movielens_explicit_df['rating'],
                                allow_missing_ids=True)


@pytest.fixture(scope='session')
def train_val_implicit_pandas_data(movielens_implicit_df):
    return train_test_split(movielens_implicit_df, test_size=0.2)


@pytest.fixture(scope='session')
def train_val_implicit_data(movielens_implicit_interactions):
    return stratified_split(
        interactions=movielens_implicit_interactions,
        val_p=0.,
        test_p=0.2,
        seed=42,
    )


@pytest.fixture(scope='session')
def train_val_explicit_data(movielens_explicit_interactions):
    return stratified_split(
        interactions=movielens_explicit_interactions,
        val_p=0.,
        test_p=0.2,
        seed=42,
    )


@pytest.fixture(scope='session')
def train_val_implicit_sample_data(movielens_implicit_interactions):
    _, train, val = random_split(
        interactions=movielens_implicit_interactions,
        val_p=0.05,
        test_p=0.01,
        seed=42,
    )

    return train, val


@pytest.fixture(scope='session')
def train_val_explicit_sample_data(movielens_explicit_interactions):
    _, train, val = random_split(
        interactions=movielens_explicit_interactions,
        val_p=0.05,
        test_p=0.01,
        seed=42,
    )

    return train, val



from functools import partial
from typing import Callable, Dict, Optional, Tuple, Union

import torch
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings


class MatrixFactorizationModel(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for the matrix factorization model.

    ``MatrixFactorizationModel`` models have an embedding layer for both users and items which are
    dot-producted together to output a single float ranking value.

    Collie adds a twist on to this incredibly popular framework by allowing separate optimizers
    for embeddings and bias terms. With larger datasets and multiple epochs of training, a model
    might incorrectly learn to only optimize the bias terms for a quicker path towards a local
    loss minimum, essentially memorizing how popular each item is. By using a separate, slower
    optimizer for the bias terms (like Stochastic Gradient Descent), the model must prioritize
    optimizing the embeddings for meaningful, more varied recommendations, leading to a model
    that is able to achieve a much lower loss. See the documentation below for ``bias_lr`` and
    ``bias_optimizer`` input arguments for implementation details.

    All ``MatrixFactorizationModel`` instances are subclasses of the ``LightningModule`` class
    provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, MatrixFactorizationModel


        model = MatrixFactorizationModel(train=train)
        trainer = CollieTrainer(model)
        trainer.fit(model)
        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(filename='model.pth')
        new_model = MatrixFactorizationModel(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Parameters
    ----------
    embedding_dim: int
        Number of latent factors to use for user and item embeddings
    dropout_p: float
        Probability of dropout
    sparse: bool
        Whether or not to treat embeddings as sparse tensors. If ``True``, cannot use weight decay
        on the optimizer
    bias_lr: float
        Bias terms learning rate. If 'infer', will set equal to ``lr``
    bias_optimizer: torch.optim or str
        Optimizer for the bias terms. This supports the same string options as ``optimizer``, with
        the addition of ``infer``, which will set the optimizer equal to ``optimizer``. If
        ``bias_optimizer`` is ``None``, only a single optimizer will be created for all model
        parameters
    y_range: tuple
        Specify as ``(min, max)`` to apply a sigmoid layer to the output score of the model to get
        predicted ratings within the range of ``min`` and ``max``

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 embedding_dim: int = 30,
                 dropout_p: float = 0.0,
                 sparse: bool = False,
                 lr: float = 1e-3,
                 bias_lr: Optional[Union[float, str]] = 1e-2,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 bias_optimizer: Optional[Union[str, Callable]] = 'sgd',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        super().__init__(**get_init_arguments())

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after ``prepare_data``.

        """
        self.user_biases = ZeroEmbedding(num_embeddings=self.hparams.num_users,
                                         embedding_dim=1,
                                         sparse=self.hparams.sparse)
        self.item_biases = ZeroEmbedding(num_embeddings=self.hparams.num_items,
                                         embedding_dim=1,
                                         sparse=self.hparams.sparse)
        self.user_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                               embedding_dim=self.hparams.embedding_dim,
                                               sparse=self.hparams.sparse)
        self.item_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                               embedding_dim=self.hparams.embedding_dim,
                                               sparse=self.hparams.sparse)
        self.dropout = nn.Dropout(p=self.hparams.dropout_p)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Simple matrix factorization for a single user and item looks like:

        ````prediction = (user_embedding * item_embedding) + user_bias + item_bias````

        If dropout is added, it is applied to the two embeddings and not the biases.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embeddings = self.user_embeddings(users)
        item_embeddings = self.item_embeddings(items)

        preds = (
            torch.mul(self.dropout(user_embeddings), self.dropout(item_embeddings)).sum(axis=1)
            + self.user_biases(users).squeeze(1)
            + self.item_biases(items).squeeze(1)
        )

        if self.hparams.y_range is not None:
            preds = (
                torch.sigmoid(preds)
                * (self.hparams.y_range[1] - self.hparams.y_range[0])
                + self.hparams.y_range[0]
            )

        return preds

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        return self.item_embeddings.weight.data



from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from pathlib import Path
import textwrap
from typing import Callable, Dict, List, Optional, Tuple, Union
import warnings

import numpy as np
import pandas as pd
from pytorch_lightning.core.lightning import LightningModule
import torch

from collie.interactions import (ApproximateNegativeSamplingInteractionsDataLoader,
                                 ExplicitInteractions,
                                 Interactions,
                                 InteractionsDataLoader)
from collie.loss import (adaptive_bpr_loss,
                         adaptive_hinge_loss,
                         bpr_loss,
                         hinge_loss,
                         warp_loss)
from collie.utils import get_init_arguments


INTERACTIONS_LIKE_INPUT = Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                ExplicitInteractions,
                                Interactions,
                                InteractionsDataLoader]
EXPECTED_BATCH_TYPE = Union[Tuple[Tuple[torch.tensor, torch.tensor], torch.tensor],
                            Tuple[torch.tensor, torch.tensor, torch.tensor]]


class BasePipeline(LightningModule, metaclass=ABCMeta):
    """
    Base Pipeline model architectures to inherit from.

    All subclasses MUST at least override the following methods:

    * ``_setup_model`` - Set up the model architecture

    * ``forward`` - Forward pass through a model

    For ``item_item_similarity`` to work properly, all subclasses are should also implement:

    * ``_get_item_embeddings`` - Returns item embeddings from the model on the device

    Parameters
    ----------
    train: ``collie.interactions`` object
        Data loader for training data. If an ``Interactions`` object is supplied, an
        ``InteractionsDataLoader`` will automatically be instantiated with ``shuffle=True``
    val: ``collie.interactions`` object
        Data loader for validation data. If an ``Interactions`` object is supplied, an
        ``InteractionsDataLoader`` will automatically be instantiated with ``shuffle=False``
    lr: float
        Model learning rate
    lr_scheduler_func: torch.optim.lr_scheduler
        Learning rate scheduler to use during fitting
    weight_decay: float
        Weight decay passed to the optimizer, if optimizer permits
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

        * ``'sparse_adam'`` (for ``torch.optim.SparseAdam``)

    loss: function or str
        If a string, one of the following implemented losses:

        * ``'bpr'`` / ``'adaptive_bpr'`` (implicit data)

        * ``'hinge'`` / ``'adaptive_hinge'`` (implicit data)

        * ``'warp'`` (implicit data)

        * ``'mse'`` (explicit data)

        * ``'mae'`` (explicit data)

        For implicit data, if ``train.num_negative_samples > 1``, the adaptive loss version will
        automatically be used of the losses above (except for WARP loss, which is only adaptive by
        nature).

        If a callable is passed, that function will be used for calculating the loss. For implicit
        models, the first two arguments passed will be the positive and negative predictions,
        respectively. Additional keyword arguments passed in order are ``num_items``,
        ``positive_items``, ``negative_items``, ``metadata``, and ``metadata_weights``.
        For explicit models, the only two arguments passed in will be the prediction and actual
        rating values, in order.
    metadata_for_loss: dict
        Keys should be strings identifying each metadata type that match keys in
        ``metadata_weights``. Values should be a ``torch.tensor`` of shape (num_items x 1). Each
        tensor should contain categorical metadata information about items (e.g. a number
        representing the genre of the item)
    metadata_for_loss_weights: dict
        Keys should be strings identifying each metadata type that match keys in ``metadata``.
        Values should be the amount of weight to place on a match of that type of metadata, with
        the sum of all values ``<= 1``.
        e.g. If ``metadata_for_loss_weights = {'genre': .3, 'director': .2}``, then an item is:

        * a 100% match if it's the same item,

        * a 50% match if it's a different item with the same genre and same director,

        * a 30% match if it's a different item with the same genre and different director,

        * a 20% match if it's a different item with a different genre and same director,

        * a 0% match if it's a different item with a different genre and different director,
          which is equivalent to the loss without any partial credit
    load_model_path: str or Path
        To load a previously-saved model for inference, pass in path to output of
        ``model.save_model()`` method. Note that datasets and optimizers will NOT be restored.
        If ``None``, will initialize model as normal
    map_location: str or torch.device
        If ``load_model_path`` is provided, device specifying how to remap storage locations when
        ``torch.load``-ing the state dictionary
    **kwargs: keyword arguments
        All keyword arguments will be saved to ``self.hparams`` by default

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 lr: float = 1e-3,
                 lr_scheduler_func: Optional[Callable] = None,
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None,
                 **kwargs):
        if isinstance(train, Interactions) or isinstance(train, ExplicitInteractions):
            train = InteractionsDataLoader(interactions=train, shuffle=True)
        if isinstance(val, Interactions) or isinstance(val, ExplicitInteractions):
            val = InteractionsDataLoader(interactions=val, shuffle=False)

        super().__init__()

        # save datasets as class-level attributes and NOT ``hparams`` so model checkpointing /
        # saving can complete faster
        self.train_loader = train
        self.val_loader = val

        # potential issue with PyTorch Lightning is that a function cannot be saved as a
        # hyperparameter, so we will sidestep this by setting it as a class-level attribute
        # https://github.com/PyTorchLightning/pytorch-lightning/issues/2444
        self.lr_scheduler_func = lr_scheduler_func
        self.loss = loss
        self.optimizer = optimizer
        self.bias_optimizer = kwargs.get('bias_optimizer')

        if load_model_path is not None:
            # we are loading in a previously-saved model, not creating a new one
            self._load_model_init_helper(load_model_path=load_model_path,
                                         map_location=map_location,
                                         **kwargs)
        else:
            if self.train_loader is None:
                raise TypeError('``train`` must be provided to all newly-instantiated models!')
            elif self.val_loader is not None:
                assert self.train_loader.num_users == self.val_loader.num_users, (
                    'Both training and val ``num_users`` must equal: '
                    f'{self.train_loader.num_users} != {self.val_loader.num_users}.'
                )
                assert self.train_loader.num_items == self.val_loader.num_items, (
                    'Both training and val ``num_items`` must equal: '
                    f'{self.train_loader.num_items} != {self.val_loader.num_items}.'
                )

                if (
                    hasattr(self.train_loader, 'num_negative_samples')
                    or hasattr(self.val_loader, 'num_negative_samples')
                ):
                    num_negative_samples_error = (
                        'Training and val ``num_negative_samples`` property must both equal ``1``'
                        f' or both be greater than ``1``, not: {self.train_loader.num_items} and'
                        f' {self.val_loader.num_items}, respectively.'
                    )
                    if self.train_loader.num_negative_samples == 1:
                        assert self.val_loader.num_negative_samples == 1, num_negative_samples_error
                    elif self.train_loader.num_negative_samples > 1:
                        assert self.val_loader.num_negative_samples > 1, num_negative_samples_error
                    else:
                        raise ValueError(
                            '``self.train_loader.num_negative_samples`` must be greater than ``0``,'
                            f' not {self.train_loader.num_negative_samples}.'
                        )

            # saves all passed-in parameters
            init_args = get_init_arguments(
                exclude=['train', 'val', 'item_metadata', 'trained_model'],
                verbose=False,
            )

            self.save_hyperparameters(init_args, *kwargs.keys())

            self.hparams.num_users = self.train_loader.num_users
            self.hparams.num_items = self.train_loader.num_items
            self.hparams.num_epochs_completed = 0

            self._configure_loss()

            # check weight decay and sparsity
            if hasattr(self.hparams, 'sparse'):
                if self.hparams.sparse and self.hparams.weight_decay != 0:
                    warnings.warn(
                        textwrap.dedent(
                            f'''
                            ``weight_decay`` value must be 0 when ``sparse`` is flagged, not
                            {self.hparams.weight_decay}. Setting to 0.
                            '''
                        ).replace('\n', ' ').strip()
                    )
                    self.hparams.weight_decay = 0.0

            # set up the actual model
            self._setup_model(**kwargs)

    def _load_model_init_helper(self, load_model_path: str, map_location: str, **kwargs) -> None:
        loaded_dict = torch.load(str(load_model_path), map_location=map_location)

        for key, value in loaded_dict['hparams'].items():
            self.hparams[key] = value

        self.hparams['load_model_path'] = load_model_path
        self.hparams['map_location'] = map_location

        self._setup_model(**kwargs)

        self.load_state_dict(state_dict=loaded_dict['state_dict'])
        self.eval()

    @abstractmethod
    def _setup_model(self, **kwargs) -> None:
        """``_setup_model`` should be implemented in all subclasses."""
        pass

    @abstractmethod
    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """``forward`` should be implemented in all subclasses."""
        pass

    def _move_any_external_data_to_device(self):
        """Code for ensuring all side-data is put onto the model's device before training."""
        pass

    def _configure_loss(self) -> None:
        # set up loss function
        self.loss_function = None

        if callable(self.loss):
            self.loss_function = self.loss
            return

        # explicit losses first
        self.hparams._is_implicit = False
        if self.loss == 'mse':
            self.loss_function = torch.nn.MSELoss(reduction='mean')
            return
        elif self.loss == 'mae':
            self.loss_function = torch.nn.L1Loss(reduction='mean')
            return

        # followed by implicit losses
        self.hparams._is_implicit = True
        if not hasattr(self.train_loader, 'num_negative_samples'):
            raise ValueError(
                '``num_negative_samples`` attribute not found in ``train_loader`` - are you using '
                'explicit data with an implicit loss function?'
            )
        elif self.loss == 'warp':
            if self.train_loader.num_negative_samples > 1:
                self.loss_function = warp_loss
                return
            else:
                raise ValueError('Cannot use WARP loss with a single negative sample!')
        elif 'bpr' in self.loss:
            if self.train_loader.num_negative_samples > 1:
                self.loss_function = adaptive_bpr_loss
            else:
                if 'adaptive' in self.loss:
                    warnings.warn(
                        textwrap.dedent(
                            '''
                            Adaptive BPR loss specified, but ``num_negative_samples`` == 1. Using
                            standard BPR loss instead.
                            '''
                        ).replace('\n', ' ').strip()
                    )

                self.loss_function = bpr_loss
            return
        elif 'hinge' in self.loss or self.loss == 'adaptive':
            if self.train_loader.num_negative_samples > 1:
                self.loss_function = adaptive_hinge_loss
            else:
                if 'adaptive' in self.loss:
                    warnings.warn(
                        textwrap.dedent(
                            '''
                            Adaptive hinge loss specified, but ``num_negative_samples`` == 1. Using
                            standard hinge loss instead.
                            '''
                        ).replace('\n', ' ').strip()
                    )

                self.loss_function = hinge_loss
            return
        else:
            raise ValueError('{} is not a valid loss function.'.format(self.loss))

    def configure_optimizers(self) -> (
        Union[Tuple[List[Callable], List[Callable]], Tuple[Callable, Callable], Callable]
    ):
        """
        Configure optimizers and learning rate schedulers to use in optimization.

        This method will be called after ``setup``.

        If ``self.bias_optimizer`` is None, only a single optimizer will be returned. If there is a
        non-``None`` class attribute for ``bias_optimizer``, two optimizers will be created: one
        for all layers with the name 'bias' in them, and another for all other model parameters.
        The bias optimizer will be set with the same parameters as ``optimizer`` with the
        exception of the learning rate, which will be set to ``self.hparams.bias_lr``.

        """
        # since this is the only function that is called before each ``trainer.fit`` call, we will
        # also take this time to ensure any external data a model might rely on has been properly
        # moved to the device before training
        self._move_any_external_data_to_device()

        if self.bias_optimizer is not None:
            if self.bias_optimizer == 'infer':
                self.bias_optimizer = self.optimizer
            if self.hparams.bias_lr == 'infer':
                self.hparams.bias_lr = self.hparams.lr

            # create one optimizer only for layers with the term 'bias' in them
            bias_optimizer = self._get_optimizer(self.bias_optimizer, optimizer_type='bias')
            # create another optimizer for all other model layers
            optimizer = self._get_optimizer(self.optimizer, optimizer_type='all_but_bias')
        else:
            # create a single optimizer for all model layers
            optimizer = self._get_optimizer(self.optimizer, optimizer_type='all')

        if self.lr_scheduler_func is not None:
            monitor = 'val_loss_epoch'
            if self.val_loader is None:
                monitor = 'train_loss_epoch'

            # add in optimizer to scheduler function
            scheduler_dict = {
                'scheduler': self.lr_scheduler_func(optimizer),
                'monitor': monitor,
            }

            optimizer_list = [optimizer]
            scheduler_list = [scheduler_dict]

            # create a separate learning rate scheduler for our bias optimizer if it exists
            if self.bias_optimizer is not None:
                scheduler_dict_bias = {
                    'scheduler': self.lr_scheduler_func(bias_optimizer),
                    'monitor': monitor,
                }

                optimizer_list.append(bias_optimizer)
                scheduler_list.append(scheduler_dict_bias)

            return (optimizer_list, scheduler_list)

        if self.bias_optimizer is not None:
            return [optimizer, bias_optimizer]
        else:
            return optimizer

    def _get_optimizer(self, optimizer: Optional[Union[str, Callable]], **kwargs) -> Callable:
        if callable(optimizer):
            try:
                optimizer = optimizer(
                    self._get_optimizer_parameters(include_weight_decay=True, **kwargs)
                )
            except TypeError:
                optimizer = optimizer(
                    self._get_optimizer_parameters(include_weight_decay=False, **kwargs)
                )
        elif optimizer == 'sgd':
            optimizer = torch.optim.SGD(
                self._get_optimizer_parameters(include_weight_decay=True, **kwargs)
            )
        elif optimizer == 'adam':
            optimizer = torch.optim.Adam(
                self._get_optimizer_parameters(include_weight_decay=True, **kwargs)
            )
        elif optimizer == 'sparse_adam':
            optimizer = torch.optim.SparseAdam(
                self._get_optimizer_parameters(include_weight_decay=False, **kwargs)
            )
        else:
            raise ValueError('{} is not a valid optimizer!'.format(optimizer))

        return optimizer

    def _get_optimizer_parameters(self,
                                  include_weight_decay: bool = True,
                                  optimizer_type: str = 'all',
                                  **kwargs) -> Dict[str, Union[torch.tensor, float]]:
        """
        Set all non-bias model layers with ``lr`` learning rate and bias terms with ``bias_lr``
        learning rate.

        """
        assert optimizer_type in ['bias', 'all_but_bias', 'all'], f'{optimizer_type} not valid!'

        if optimizer_type == 'bias':
            optimizer_parameters = [
                {
                    'params': (
                        param for (name, param) in self.named_parameters() if 'bias' in name
                    ),
                    'lr': self.hparams.bias_lr,
                },
            ]
        elif optimizer_type == 'all_but_bias':
            optimizer_parameters = [
                {
                    'params': (
                        param for (name, param) in self.named_parameters() if 'bias' not in name
                    ),
                    'lr': self.hparams.lr,
                },
            ]
        else:
            optimizer_parameters = [{
                'params': self.parameters(),
                'lr': self.hparams.lr,
            }]

        if include_weight_decay:
            weight_decay_dict = {'weight_decay': self.hparams.weight_decay}
            for d in optimizer_parameters:
                d.update(weight_decay_dict)

        return optimizer_parameters

    def train_dataloader(self) -> Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                        InteractionsDataLoader]:
        """
        Method that sets up training data as a PyTorch DataLoader.

        This method will be called after ``configure_optimizers``.

        """
        return self.train_loader

    def training_step(
        self,
        batch: Tuple[Tuple[torch.tensor, torch.tensor], torch.tensor],
        batch_idx: int,
        optimizer_idx: Optional[int] = None,
    ) -> torch.tensor:
        """
        Method that contains logic for what happens inside the training loop.

        This method will be called after ``train_dataloader``.

        """
        loss = self.calculate_loss(batch)

        # add logging
        self.log(name='train_loss_step', value=loss)

        return loss

    def training_epoch_end(
        self,
        outputs: Union[List[float], List[List[float]]],
    ) -> None:
        """
        Method that contains a callback for logic to run after the training epoch ends.

        This method will be called after ``training_step``.

        """
        self.hparams.num_epochs_completed += 1

        try:
            avg_loss = torch.stack([x['loss'] for x in outputs]).mean()
        except TypeError:
            # with multiple optimizers, ``outputs`` has a list within a list
            avg_loss = torch.stack([x['loss'] for x in outputs[0]]).mean()

        self.log(name='train_loss_epoch', value=avg_loss)

    def val_dataloader(self) -> Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                      InteractionsDataLoader]:
        """
        Method that sets up validation data as a PyTorch DataLoader.

        This method will be called after ``training_step``.

        """
        return self.val_loader

    def validation_step(
        self,
        batch: Tuple[Tuple[torch.tensor, torch.tensor], torch.tensor],
        batch_idx: int,
        optimizer_idx: Optional[int] = None,
    ) -> torch.tensor:
        """
        Method that contains logic for what happens inside the validation loop.

        This method will be called after ``val_dataloader``.

        """
        loss = self.calculate_loss(batch)

        # add logging
        self.log(name='val_loss_step', value=loss)

        return loss

    def validation_epoch_end(
        self,
        outputs: List[float]
    ) -> None:
        """
        Method that contains a callback for logic to run after the validation epoch ends.

        This method will be called after ``validation_step``.

        """
        avg_loss = torch.stack([x for x in outputs]).mean()

        self.log(name='val_loss_epoch', value=avg_loss)

    def calculate_loss(self, batch: EXPECTED_BATCH_TYPE) -> torch.tensor:
        """
        Given a batch of data, calculate the loss value.

        Note that the data type (implicit or explicit) will be determined by the structure of the
        batch sent to this method. See the table below for expected data types:

        .. list-table::
            :header-rows: 1

            * - ``__getitem__`` Format
              - Expected Meaning
              - Model Type
            * - ``((X, Y), Z)``
              - ``((user IDs, item IDs), negative item IDs)``
              - **Implicit**
            * - ``(X, Y, Z)``
              - ``(user IDs, item IDs, ratings)``
              - **Explicit**

        """
        if len(batch) == 2 and isinstance(batch[0], Iterable) and len(batch[0]) == 2:
            if self.hparams.get('_is_implicit') is False:
                raise ValueError('Explicit loss with implicit data is invalid!')

            # implicit data
            ((users, pos_items), neg_items) = batch

            users = users.long()
            pos_items = pos_items.long()
            # TODO: see if there is a way to not have to transpose each time - probably a bit costly
            neg_items = torch.transpose(neg_items, 0, 1).long()

            # get positive item predictions from model
            pos_preds = self(users, pos_items)

            # get negative item predictions from model
            users_repeated = users.repeat(neg_items.shape[0])
            neg_items_flattened = neg_items.flatten()
            neg_preds = self(users_repeated, neg_items_flattened).view(
                neg_items.shape[0], len(users)
            )

            # implicit loss function
            loss = self.loss_function(
                pos_preds,
                neg_preds,
                num_items=self.hparams.num_items,
                positive_items=pos_items,
                negative_items=neg_items,
                metadata=self.hparams.metadata_for_loss,
                metadata_weights=self.hparams.metadata_for_loss_weights,
            )
        elif len(batch) == 3:
            if self.hparams.get('_is_implicit') is True:
                raise ValueError('Implicit loss with explicit data is invalid!')

            # explicit data
            (users, items, ratings) = batch

            users = users.long()
            items = items.long()
            ratings = ratings.float()

            # get predictions from model
            preds = self(users, items)

            # explicit loss function
            loss = self.loss_function(preds, ratings)
        else:
            raise ValueError(f'Unexpected format for batch: {batch}. See docs for expected format.')

        return loss

    def get_item_predictions(self,
                             user_id: int = 0,
                             unseen_items_only: bool = False,
                             sort_values: bool = True) -> pd.Series:
        """
        Get predicted rankings/ratings for all items for a given ``user_id``.

        This method cannot be called for datasets stored in ``HDF5InteractionsDataLoader`` since
        data in this ``DataLoader`` is read in dynamically.

        Parameters
        ----------
        user_id: int
        unseen_items_only: bool
            Filter ``preds`` to only show predictions of unseen items not present in the training
            or validation datasets for that ``user_id``. Note this requires both ``train_loader``
            and ``val_loader`` to be 1) class-level attributes in the model and 2) DataLoaders with
            ``Interactions`` at its core (not ``HDF5Interactions``). If you are loading in a model,
            these two attributes will need to be set manually, since datasets are NOT saved when
            saving the model
        sort_values: bool
            Whether to sort recommendations by descending prediction probability or not

        Returns
        -------
        preds: pd.Series
            Sorted values as predicted ratings for each item in the dataset with the index being
            the item ID

        """
        user = torch.tensor(
            [user_id] * self.hparams.num_items,
            dtype=torch.long,
            device=self.device
        )
        item = torch.arange(self.hparams.num_items, dtype=torch.long, device=self.device)

        preds = self(user, item)
        preds = preds.detach().cpu()
        preds = pd.Series(preds)
        if sort_values:
            preds = preds.sort_values(ascending=False)

        if unseen_items_only:
            idxs_to_drop = self.train_loader.mat.tocsr()[user_id, :].nonzero()[1]
            if self.val_loader is not None:
                idxs_to_drop = np.concatenate([
                    self.train_loader.mat.tocsr()[user_id, :].nonzero()[1],
                    self.val_loader.mat.tocsr()[user_id, :].nonzero()[1]
                ])
            filtered_preds = preds.drop(idxs_to_drop)

            return filtered_preds
        else:
            return preds

    def item_item_similarity(self, item_id: int) -> pd.Series:
        """
        Get most similar item indices by cosine similarity.

        Cosine similarity is computed with item embeddings from a trained model.

        Parameters
        ----------
        item_id: int

        Returns
        -------
        sim_score_idxs: pd.Series
            Sorted values as cosine similarity for each item in the dataset with the index being
            the item ID

        Note
        ----
        Returned array is unfiltered, so the first element, being the most similar item, will
        always be the item itself.

        """
        item_embeddings = self._get_item_embeddings()
        item_embeddings = item_embeddings / item_embeddings.norm(dim=1)[:, None]

        sim_score_idxs = (
            torch.matmul(item_embeddings[[item_id], :], item_embeddings.transpose(1, 0))
            .detach()
            .cpu()
            .numpy()
            .squeeze()
        )

        sim_score_idxs_series = pd.Series(sim_score_idxs)
        sim_score_idxs_series = sim_score_idxs_series.sort_values(ascending=False)

        return sim_score_idxs_series

    def _get_item_embeddings(self) -> torch.tensor:
        """``_get_item_embeddings`` should be implemented in all subclasses."""
        raise NotImplementedError(
            '``BasePipeline`` is meant to be inherited from, not used. '
            '``_get_item_embeddings`` is not implemented in this subclass.'
        )

    def save_model(self, filename: Union[str, Path] = 'model.pth') -> None:
        """
        Save the model's state dictionary and hyperparameters.

        While PyTorch Lightning offers a way to save and load models, there are two main reasons
        for overriding these:

        1) We only want to save the underlying PyTorch model (and not the ``Trainer`` object) so
           we don't have to require PyTorch Lightning as a dependency when deploying a model.

        2) In the v0.8.4 release, loading a model back in leads to a ``RuntimeError`` unable to
           load in weights.

        Parameters
        ----------
        filepath: str or Path
            Filepath for state dictionary to be saved at ending in '.pth'

        """
        dict_to_save = {'state_dict': self.state_dict(), 'hparams': self.hparams}
        torch.save(dict_to_save, str(filename))



from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings


class NonlinearMatrixFactorizationModel(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for a nonlinear matrix factorization model.

    ``NonlinearMatrixFactorizationModel`` models have an embedding layer for users and items. These
    are sent through separate dense networks, which output more refined embeddings, which are then
    dot producted for a single float ranking / rating.

    Collie adds a twist on to this novel framework by allowing separate optimizers for embeddings
    and bias terms. With larger datasets and multiple epochs of training, a model might incorrectly
    learn to only optimize the bias terms for a quicker path towards a local loss minimum,
    essentially memorizing how popular each item is. By using a separate, slower optimizer for the
    bias terms (like Stochastic Gradient Descent), the model must prioritize optimizing the
    embeddings for meaningful, more varied recommendations, leading to a model that is able to
    achieve a much lower loss. See the documentation below for ``bias_lr`` and ``bias_optimizer``
    input arguments for implementation details.

    All ``NonlinearMatrixFactorizationModel`` instances are subclasses of the ``LightningModule``
    class provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, NonlinearMatrixFactorizationModel


        model = NonlinearMatrixFactorizationModel(train=train)
        trainer = CollieTrainer(model)
        trainer.fit(model)
        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(filename='model.pth')
        new_model = NonlinearMatrixFactorizationModel(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Parameters
    ----------
    user_embedding_dim: int
        Number of latent factors to use for user embeddings
    item_embedding_dim: int
        Number of latent factors to use for item embeddings
    user_dense_layers_dims: list
        List of linear layer dimensions to apply to the user embedding, starting with the dimension
        directly following ``user_embedding_dim``
    item_dense_layers_dims: list
        List of linear layer dimensions to apply to the item embedding, starting with the dimension
        directly following ``item_embedding_dim``
    embedding_dropout_p: float
        Probability of dropout on the embedding layers
    dense_dropout_p: float
        Probability of dropout on the dense layers
    bias_lr: float
        Bias terms learning rate. If 'infer', will set equal to ``lr``
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    bias_optimizer: torch.optim or str
        Optimizer for the bias terms. This supports the same string options as ``optimizer``, with
        the addition of ``infer``, which will set the optimizer equal to ``optimizer``. If
        ``bias_optimizer`` is ``None``, only a single optimizer will be created for all model
        parameters
    y_range: tuple
        Specify as ``(min, max)`` to apply a sigmoid layer to the output score of the model to get
        predicted ratings within the range of ``min`` and ``max``

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 user_embedding_dim: int = 60,
                 item_embedding_dim: int = 60,
                 user_dense_layers_dims: List[float] = [48, 32],
                 item_dense_layers_dims: List[float] = [48, 32],
                 embedding_dropout_p: float = 0.0,
                 dense_dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 bias_lr: Optional[Union[float, str]] = 1e-2,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 bias_optimizer: Optional[Union[str, Callable]] = 'sgd',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        super().__init__(**get_init_arguments())

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after ``prepare_data``.

        """
        self.user_biases = ZeroEmbedding(num_embeddings=self.hparams.num_users,
                                         embedding_dim=1)
        self.item_biases = ZeroEmbedding(num_embeddings=self.hparams.num_items,
                                         embedding_dim=1)
        self.user_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                               embedding_dim=self.hparams.user_embedding_dim)
        self.item_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                               embedding_dim=self.hparams.item_embedding_dim)

        self.embedding_dropout = nn.Dropout(p=self.hparams.embedding_dropout_p)
        self.dense_dropout = nn.Dropout(p=self.hparams.dense_dropout_p)

        # set up user dense layers
        user_dense_layers_dims = (
            [self.hparams.user_embedding_dim] + self.hparams.user_dense_layers_dims
        )
        self.user_dense_layers = [
            nn.Linear(user_dense_layers_dims[idx - 1], user_dense_layers_dims[idx])
            for idx in range(1, len(user_dense_layers_dims))
        ]
        for i, layer in enumerate(self.user_dense_layers):
            nn.init.xavier_normal_(self.user_dense_layers[i].weight)
            self.add_module('user_dense_layer_{}'.format(i), layer)

        # set up item dense layers
        item_dense_layers_dims = (
            [self.hparams.item_embedding_dim] + self.hparams.item_dense_layers_dims
        )
        self.item_dense_layers = [
            nn.Linear(item_dense_layers_dims[idx - 1], item_dense_layers_dims[idx])
            for idx in range(1, len(item_dense_layers_dims))
        ]
        for i, layer in enumerate(self.item_dense_layers):
            nn.init.xavier_normal_(self.item_dense_layers[i].weight)
            self.add_module('item_dense_layer_{}'.format(i), layer)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embeddings = self.user_embeddings(users)
        item_embeddings = self.item_embeddings(items)

        for idx, user_dense_layer in enumerate(self.user_dense_layers):
            user_embeddings = F.leaky_relu(
                user_dense_layer(user_embeddings)
            )

            if idx < (len(self.user_dense_layers) - 1):
                user_embeddings = self.dense_dropout(user_embeddings)

        for idx, item_dense_layer in enumerate(self.item_dense_layers):
            item_embeddings = F.leaky_relu(
                item_dense_layer(item_embeddings)
            )

            if idx < (len(self.item_dense_layers) - 1):
                item_embeddings = self.dense_dropout(item_embeddings)

        preds = (
            (
                self.embedding_dropout(user_embeddings) * self.embedding_dropout(item_embeddings)
            ).sum(1)
            + self.user_biases(users).squeeze(1)
            + self.item_biases(items).squeeze(1)
        )

        if self.hparams.y_range is not None:
            preds = (
                torch.sigmoid(preds)
                * (self.hparams.y_range[1] - self.hparams.y_range[0])
                + self.hparams.y_range[0]
            )

        return preds

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        if not hasattr(self, 'item_embeddings_'):
            items = torch.arange(self.hparams.num_items, device=self.device)

            item_embeddings = self.item_embeddings(items)

            for item_dense_layer in self.item_dense_layers:
                item_embeddings = F.leaky_relu(
                    item_dense_layer(item_embeddings)
                )

            self.item_embeddings_ = item_embeddings.detach()

        return self.item_embeddings_
